# sdcat, Apache-2.0 license
# Filename: __main__.py
# Description: Main entry point for the sdcat command line interface
from datetime import datetime
from pathlib import Path

import click
import sys

sys.path.insert(0, str(Path(__file__).parent.parent))
from sdcat.logger import err, info, create_logger_file
from sdcat import __version__
from sdcat.cluster.commands import run_cluster_det, run_cluster_roi
from sdcat.detect.commands import run_detect


create_logger_file("sdcat")
default_data_path = Path(__file__).parent / "testdata"
default_model = "MBARI/megabenthic"


@click.group(context_settings={"help_option_names": ["-h", "--help"]})
@click.version_option(__version__, "-V", "--version", message="%(prog)s, version %(version)s")
def cli():
    """
    Process images either to detect or cluster similar objects from a command line.
    """
    pass


cli.add_command(run_detect)


@cli.group(name="cluster")
def cli_cluster():
    """
    Commands related to clustering images
    """
    pass


cli.add_command(cli_cluster)
cli_cluster.add_command(run_cluster_det)
cli_cluster.add_command(run_cluster_roi)


if __name__ == "__main__":
    try:
        start = datetime.utcnow()
        cli()
        end = datetime.utcnow()
        info(f"Done. Elapsed time: {end - start} seconds")
    except Exception as e:
        err(f"Exiting. Error: {e}")
        exit(-1)



import hashlib
import uuid

import cv2
import pandas as pd
import torch
from sahi.postprocess.combine import nms

from sdcat.cluster.utils import crop_square_image
from sdcat.logger import warn, info


def process_image(
    image_filename,
    save_path_base,
    save_path_det_raw,
    save_path_det_filtered,
    save_path_det_roi,
    save_path_viz,
    min_area,
    max_area,
    min_saliency,
    class_agnostic,
    save_roi,
    roi_size,
):
    try:
        img_color = cv2.imread(image_filename.as_posix())
        height, width = img_color.shape[:2]

        # Path to save final csv file with the detections
        pred_out_csv = save_path_det_filtered / f"{image_filename.stem}.csv"

        df_combined = pd.DataFrame()
        for csv_file in save_path_det_raw.rglob(f"{image_filename.stem}*csv"):
            df = pd.read_csv(csv_file, sep=",")
            df_combined = pd.concat([df_combined, df])

        if df_combined.empty:
            warn(f"No detections found in {image_filename}")
            return 0

        df_combined = df_combined[(df_combined["area"] > min_area) & (df_combined["area"] < max_area)]
        df_combined = df_combined[(df_combined["saliency"] > min_saliency) | (df_combined["saliency"] == -1)]

        if df_combined.empty:
            warn(f"No detections found in {image_filename}")
            return 0

        pred_list = torch.tensor(df_combined[["x", "y", "xx", "xy", "score"]].values.tolist())
        nms_pred_idx = nms(pred_list, "IOU", 0.1)

        df_final = df_combined.iloc[nms_pred_idx].reset_index(drop=True)
        df_final["saliency"] = df_combined["saliency"].iloc[nms_pred_idx].reset_index(drop=True)
        df_final["area"] = df_combined["area"].iloc[nms_pred_idx].reset_index(drop=True)

        pred_list = df_final[["x", "y", "xx", "xy", "score", "class"]].values
        for p in pred_list:
            if class_agnostic:
                img_color = cv2.rectangle(img_color, (int(p[0]), int(p[1])), (int(p[2]), int(p[3])), (81, 12, 51), 3)
                img_color = cv2.putText(
                    img_color,
                    f"{p[5]} {p[4]:.2f}",
                    (int(p[0]), int(p[1])),
                    cv2.FONT_HERSHEY_SIMPLEX,
                    1,
                    (255, 255, 255),
                    2,
                    cv2.LINE_AA,
                )
            else:
                md5_hash = hashlib.md5(p[5].encode())
                hex_color = md5_hash.hexdigest()[:6]
                r, g, b = (int(hex_color[:2], 16), int(hex_color[2:4], 16), int(hex_color[4:], 16))
                color = (r % 256, g % 256, b % 256)
                img_color = cv2.rectangle(img_color, (int(p[0]), int(p[1])), (int(p[2]), int(p[3])), color, 3)
                img_color = cv2.putText(
                    img_color, p[5], (int(p[0]), int(p[1])), cv2.FONT_HERSHEY_SIMPLEX, 1, color, 2, cv2.LINE_AA
                )

        output_path = save_path_viz / f"{image_filename.stem}.jpg"
        cv2.imwrite(str(output_path), img_color)

        df_final["cluster"] = -1
        df_final["image_path"] = image_filename.as_posix()
        df_final["image_width"] = width
        df_final["image_height"] = height
        df_final["x"] /= width
        df_final["y"] /= height
        df_final["xx"] /= width
        df_final["xy"] /= height
        df_final["w"] = df_final["xx"] - df_final["x"]
        df_final["h"] = df_final["xy"] - df_final["y"]

        if save_roi:
            # Add in a column for the unique crop name for each detection with a unique id
            # create a unique uuid based on the md5 hash of the box in the row
            df_final["crop_path"] = df_combined.apply(
                lambda row: f"{save_path_det_roi}/{uuid.uuid5(uuid.NAMESPACE_DNS, str(row['x']) + str(row['y']) + str(row['xx']) + str(row['xy']))}.png",
                axis=1,
            )
            # Crop the square image
            for index, row in df_final.iterrows():
                crop_square_image(row, roi_size)

        # Save DataFrame to CSV file including image_width and image_height columns
        info(f"Detections saved to {pred_out_csv}")
        df_final.to_csv(pred_out_csv.as_posix(), index=False, header=True)
        if save_roi:
            info(f"ROI crops saved in {save_path_det_roi}")

        save_stats = save_path_base / "stats.txt"
        with save_stats.open("w") as sf:
            sf.write(f"Statistics for {image_filename}:\n")
            sf.write("----------------------------------\n")
            sf.write(f"Total number of bounding boxes: {df_final.shape[0]}\n")
            sf.write(f"Total number of images with (bounding box) detections found: {df_final['image_path'].nunique()}\n")
            sf.write(f"Average number of bounding boxes per image: {df_final.shape[0] / df_final['image_path'].nunique()}\n")
            sf.write(f"Average width of bounding boxes: {df_final['w'].mean() * width}\n")
            sf.write(f"Average height of bounding boxes: {df_final['h'].mean() * height}\n")
            sf.write(f"Average area of bounding boxes: {df_final['area'].mean()}\n")
            sf.write(f"Average score of bounding boxes: {df_final['score'].mean()}\n")
            sf.write(f"Average saliency of bounding boxes: {df_final['saliency'].mean()}\n")
        return len(df_final)
    except Exception as e:
        warn(f"Error processing {image_filename}: {e}")
        return 0



import os
from pathlib import Path

from sahi import AutoDetectionModel
from huggingface_hub import hf_hub_download

from sdcat.logger import err


def create_model(model: str, conf: float, device: str, model_type=None):
    """
    Utility to determine the model type, model path, and create a detection model using SAHI.

    Args:
        model (str): The name of the model to use. Can be a predefined name or a local file path.
        conf (float): Confidence threshold for the model.
        device (str): The device to run the model on ('cpu', 'cuda', etc.).
        model_type (str): The type of model to use (e.g., 'yolov5', 'huggingface'). If None, will auto-detect

    Returns:
        detection_model: An instance of the AutoDetectionModel.
    """

    # Check if the provided model is a local file path
    if os.path.exists(model):
        if os.path.isdir(model):
            dir_to_model_map = {"yolov5": "yolov5", "yolov8": "yolov8", "huggingface": "huggingface"}
            model_path = [f for f in os.listdir(model) if f.endswith(".pt")]
            if len(model_path) == 0:
                err(f"No .pt file found in directory: {model}")
                raise ValueError(f"No .pt file found in directory: {model}")
            if model_type is None:
                for k, v in dir_to_model_map.items():
                    if k in model:
                        model_type = v
                        break
            if model_type is None:
                err(
                    f"Could not determine model type from directory name: {model}. Try the --model-type option, e.g., --model-type yolov11"
                )
                raise ValueError(
                    f"Could not determine model type from directory name: {model}. Try the --model-type option, e.g., --model-type yolov11"
                )
            detection_model = AutoDetectionModel.from_pretrained(
                model_type=model_type,
                model_path=Path(model) / model_path[0],
                confidence_threshold=conf,
                device=device,
            )
            return detection_model
        else:
            raise ValueError(f"Model path is not a directory: {model}")

    # Predefined model mapping
    model_map = {
        "yolov8s": {"model_type": "yolov8", "model_path": "ultralyticsplus/yolov8s"},
        "yolov8x": {"model_type": "yolov8", "model_path": "yolov8x.pt"},
        "hustvl/yolos-small": {
            "model_type": "huggingface",
            "model_path": "hustvl/yolos-small",
            "config_path": "hustvl/yolos-small",
        },
        "hustvl/yolos-tiny": {
            "model_type": "huggingface",
            "model_path": "hustvl/yolos-tiny",
            "config_path": "hustvl/yolos-tiny",
        },
        "MBARI-org/megamidwater": {
            "model_type": "yolov5",
            "model_path": lambda: hf_hub_download("MBARI-org/megamidwater", "best.pt"),
        },
        "MBARI-org/uav-yolov5-30k": {
            "model_type": "yolov5",
            "model_path": lambda: hf_hub_download("MBARI-org/yolov5x6-uav-30k", "yolov5x6-uav-30k.pt"),
        },
        "MBARI-org/uav-yolov5-18k": {
            "model_type": "yolov5",
            "model_path": lambda: hf_hub_download("MBARI-org/yolov5-uav-18k", "yolov5x6-uav-18k.pt"),
        },
        "MBARI-org/yolo11x-uavs-detect": {
            "model_type": "yolo11",
            "model_path": lambda: hf_hub_download("MBARI-org/yolo11x-uavs-detect", "uavs-oneclass-best.pt"),
        },
        "FathomNet/MBARI-315k-yolov5": {
            "model_type": "yolov5",
            "model_path": lambda: hf_hub_download("FathomNet/MBARI-315k-yolov5", "mbari_315k_yolov5.pt"),
        },
    }

    if model not in model_map:
        raise ValueError(
            f"Unknown model: {model}. Available models: {list(model_map.keys())}, "
            f"or provide a local file path. You can also use the --model-type option to specify the model type."
        )

    model_info = model_map[model]
    model_type = model_info["model_type"]
    model_path = model_info["model_path"]

    if callable(model_path):  # If the path is a function (e.g., requires download)
        model_path = model_path()

    config_path = model_info.get("config_path", None)

    detection_model = AutoDetectionModel.from_pretrained(
        model_type=model_type,
        model_path=model_path,
        config_path=config_path,
        confidence_threshold=conf,
        device=device,
    )

    return detection_model



# sdcat, Apache-2.0 license
# Filename: sdcat/cluster/cluster.py
# Description: Clustering using vision transformer features and HDBSCAN density-based clustering
import uuid
import warnings
from importlib.util import find_spec

from pathlib import Path
from concurrent.futures import ThreadPoolExecutor, ProcessPoolExecutor, as_completed
import os
from typing import List

import pandas
import seaborn as sns
import modin.pandas as pd
import numpy as np
from sklearn.decomposition import PCA
from sklearn.preprocessing import normalize
from tqdm import tqdm
from umap import UMAP
from hdbscan import HDBSCAN
from scipy.cluster.hierarchy import linkage, fcluster
from sklearn.metrics.pairwise import cosine_similarity
from sdcat.logger import info, warn, debug
from sdcat.cluster.utils import cluster_grid, clean_bad_images, crop_all_square_images
from sdcat.cluster.embedding import fetch_embedding, has_cached_embedding, compute_norm_embedding
from sdcat import __version__ as sdcat_version

if find_spec("cuml"):
    info("=======> USING GPU for HDBSCAN AND UMAP <=========")
    from cuml.cluster import HDBSCAN as cuHDBSCAN  # pylint: disable=E0611, E0401
    from cuml.manifold.umap import UMAP as cuUMAP

    have_gpu = True
else:
    have_gpu = False

sns.set_style("darkgrid", {"axes.facecolor": ".9"})
sns.set(rc={"figure.figsize": (12, 10)})


def _summarize_clusters(
    df: pd.DataFrame,
    output_path: Path,
    prefix: str,
    cluster_sim: list,
    model: str,
    skip_visualization: bool,
    hdbscan_params: dict[str],
) -> dict:
    """
    Summarize and optionally visualize the clusters using t-SNE or UMAP in grids
    """
    info("Summarizing clusters")
    num_samples = len(df)
    df = df._to_pandas()
    clustered_df = df[df["cluster"] != -1]
    noise_df = df[df["cluster"] == -1]
    labels = clustered_df["cluster"].values
    clusters = np.unique(labels)
    num_clusters = len(clusters)
    num_labels = len(labels)

    # Get the top 20 predicted classes in percent
    info("Getting top 20 predicted classes...")

    top20 = df["class"].value_counts(normalize=True).head(20) * 100
    top20 = top20.sort_index().round(2)

    summary = {
        "dataset": {
            "output": str(output_path),
            "clustering_algorithm": "HDBSCAN",
            "clustering_parameters": hdbscan_params,
            "feature_embedding_model": str(model),
        },
        "statistics": {
            "total_clusters": num_clusters,
            "cluster_coverage": f"{float(num_labels / num_samples):.2f} ({100*float(num_labels / num_samples):.2f}%)",
        },
    }

    image_paths = {c: df.loc[df["cluster"] == c, "crop_path"] for c in clusters}

    summary["statistics"]["top_predictions"] = [
        {"class": name, "percentage": f"{total:.2f}%"} for name, total in top20.items()
    ]

    if not skip_visualization:
        # Cannot use init='spectral' when n_components is >= num_samples - default to 'random' instead
        n_components = min(2, num_samples)
        if n_components >= num_samples:
            init = "random"
        else:
            init = "spectral"

        # Reduce the dimensionality of the embeddings using UMAP to 2 dimensions to visualize the clusters
        # Only use the exemplars and a random sample of 5000 images to speed up the visualization
        sampled_df = clustered_df.sample(n=min(num_labels, 5000), random_state=42, replace=False)
        sampled_emb = [fetch_embedding(model, filename)[0] for filename in sampled_df["crop_path"]]
        np_data = np.array(sampled_emb)

        n_neighbors = min(15, num_samples - 1)
        info(f"Using {n_neighbors} neighbors for dimensional reduction")
        if n_neighbors < 2:
            warn("Using PCA instead of UMAP to reduce for cluster 2D plot")
            from sklearn.decomposition import PCA

            pca = PCA(n_components=2)
            xx = pca.fit_transform(np_data)
        else:
            if have_gpu:
                info("Using GPU to reduce cluster 2D plot")
                xx = cuUMAP(
                    init=init,
                    n_components=2,
                    n_neighbors=n_neighbors,
                    min_dist=0.1,
                    metric="euclidean",
                    verbose=False,  # print statements from the verbose flag are forcing an exit in rapids so disable it
                ).fit_transform(np_data)
            else:
                info("Using UMAP to reduce cluster 2D plot")
                xx = UMAP(init=init, n_components=2, n_neighbors=n_neighbors, metric="cosine").fit_transform(np_data)

        df_joint = pandas.DataFrame({"x": xx[:, 0], "y": xx[:, 1], "labels": sampled_df["cluster"].values})
        p = sns.jointplot(data=df_joint, x="x", y="y", hue="labels")
        p.fig.suptitle(
            f"{prefix}\nsdcat_version {sdcat_version}\nClusters {num_clusters} with {num_samples} samples", fontsize=14
        )
        p.fig.subplots_adjust(top=0.80)
        p.savefig(f"{output_path}/{prefix}_summary.png")
        info(f"Saved {output_path}/{prefix}_summary.png")

        # Create a grid of the images to check the quality of the clustering results
        num_processes = min(os.cpu_count(), num_clusters)
        info(f"Using {num_processes} CPUs to visualize the {num_clusters} clusters")

        # Use a pool of processes to speed up the visualization of the clusters
        # Skip modin here because it does not offer much speedup
        with ProcessPoolExecutor(max_workers=num_processes) as executor:
            futures = []
            for c in image_paths:
                grid_size = 4 if len(image_paths[c]) < 50 else 8  # grid size; larger clusters get larger grids
                futures.append(
                    executor.submit(cluster_grid, prefix, cluster_sim[c], c, grid_size, image_paths[c], output_path / prefix)
                )
            for future in as_completed(futures):
                future.result()

        # Create a grid of the noise images last with a different prefix
        if len(noise_df) > 0:
            grid_size = 4 if len(noise_df) < 50 else 8
            cluster_grid(prefix, 0, -1, grid_size, noise_df["crop_path"], output_path / prefix)

    return summary


def _reassign_noise(df: pd.DataFrame, exemplar_emb: np.ndarray, cluster: List[int], min_similarity: float, model: str):
    """
    Reassign the noise (-1) cluster to the nearest exemplar cluster based on cosine similarity.
    """
    noise_df = df[df["cluster"] == -1]
    noise_embeddings = np.array([fetch_embedding(model, path)[0] for path in noise_df["crop_path"]])
    similarities = cosine_similarity(noise_embeddings, exemplar_emb)
    max_scores = similarities.max(axis=1)
    best_sim_idx = similarities.argmax(axis=1)
    valid = max_scores > min_similarity
    valid_indices = noise_df.index[valid]
    df.loc[valid_indices, "cluster"] = cluster[best_sim_idx[valid]]
    return df


def _similarity_merge(
    df: pd.DataFrame, exemplar_emb: np.ndarray, cluster: List[int], min_similarity: float, model: str
) -> (pd.DataFrame, dict):
    """
    Merge clusters based on the linkage of the cosine similarity of their embeddings.
    """
    unique_clusters_before = df["cluster"].unique()

    info(f"Merging clusters with similarity threshold {min_similarity:.2f} ...")
    info(
        f"Maximum cluster id: {df['cluster'].values.max()} minimum cluster id: {df['cluster'].values.min()} unique clusters: {len(unique_clusters_before)}"
    )
    linkage_matrix = linkage(np.array(exemplar_emb), method="complete", metric="cosine")
    linkage_clusters = fcluster(linkage_matrix, 1 - min_similarity, criterion="distance")

    info(f"Unique clusters before merging: {len(unique_clusters_before)}")
    info(f"Linkage matrix size: {len(linkage_matrix)}")

    # If the cluster labels are all the same, then we have a single cluster and we can't merge
    if len(np.unique(linkage_clusters)) == 1:
        info("No clusters to merge")
    else:

        def map_to_new_cluster(old_cluster):
            if old_cluster == -1:
                debug(f"Cluster {old_cluster} is noise, skipping")
                return -1
            new_cluster = linkage_clusters[old_cluster]
            if new_cluster > len(cluster) - 1:
                return old_cluster
            actual_cluster = cluster[new_cluster]
            debug(f"Assigning cluster {old_cluster} to {actual_cluster}")
            return actual_cluster

        df["cluster"] = df["cluster"].map(map_to_new_cluster)
        unique_clusters_after = df["cluster"].unique()
        info(f"Unique clusters after merging: {len(unique_clusters_after)}")

    return df


def _compute_exemplars(df: pd.DataFrame, model: str, device: str, vits_batch_size: int) -> pd.DataFrame:
    non_noise_df = df[df["cluster"] != -1]
    max_score_idx = non_noise_df.sort_values("cluster", ascending=True).groupby("cluster")["HDBSCAN_probability"].idxmax()

    # Create a new dataframe with the exemplars for each cluster
    df_exemplars = df.loc[
        max_score_idx, ["cluster", "x", "y", "xx", "xy", "image_width", "image_height", "crop_path", "image_path"]
    ].copy()
    df_exemplars = df_exemplars.sort_values("cluster", ascending=True)
    df_exemplars["angle"] = 0

    # Add a new row for every row for every angle to be rotated
    df_exemplars_final = pd.DataFrame()
    for angle in [45, 135, 180, 225, 270, 315]:
        new_rows = df_exemplars.copy()
        new_rows["crop_path"] = new_rows["crop_path"].str.replace(".png", f"_rot{angle}.png")
        new_rows["angle"] = angle
        df_exemplars_final = pd.concat([df_exemplars_final, new_rows], ignore_index=True)

    # Add the original exemplars without rotation
    df_exemplars_final = pd.concat([df_exemplars_final, df_exemplars], ignore_index=True)

    # Crop all exemplar images at rotated angles, skipping those that already exist
    info(f"Cropping {len(df_exemplars_final)} exemplar images to square...")
    existing = df_exemplars_final["crop_path"].apply(lambda x: os.path.exists(x))
    rows_to_crop = df_exemplars_final[~existing]._to_pandas()
    image_angle_group = rows_to_crop.groupby(["image_path", "angle"])
    with ThreadPoolExecutor(max_workers=os.cpu_count()) as executor:
        futures = [
            executor.submit(crop_all_square_images, image_angle[0], df, 224, image_angle[1])
            for image_angle, df in image_angle_group
        ]
        for _ in tqdm(as_completed(futures), total=len(futures)):
            pass

    crop_paths = df_exemplars_final["crop_path"].values.tolist()
    compute_norm_embedding(model_name=model, images=crop_paths, device=device, batch_size=vits_batch_size)

    info("Computing mean embeddings for exemplars across all rotations ...")
    df_exemplar_group = df_exemplars_final.groupby("cluster")

    def calc_mean(group):
        embeddings = [fetch_embedding(model, path)[0] for path in group["crop_path"]]
        embeddings = [emb for emb in embeddings if len(emb) > 0]
        if not embeddings:
            raise RuntimeError(f"No embeddings found for {group['crop_path'].values}")
        return np.mean(np.array(embeddings), axis=0)

    # Apply the function to each group and get the mean embeddings array for the merge
    exemplar_means = df_exemplar_group.apply(calc_mean)
    exemplar_emb = np.stack(exemplar_means.values, axis=0)

    cluster = df_exemplar_group["cluster"].first().values
    return exemplar_emb, cluster


def _compute_avg_sim(df: pd.DataFrame, model: str) -> (pd.DataFrame, dict):
    def compute_cluster_avg_similarity(cluster_df):
        if len(cluster_df) == 0:
            return None

        cluster_emb = [fetch_embedding(model, f)[0] for f in cluster_df["crop_path"]]
        cluster_emb = np.array(cluster_emb)

        # This computes self-similarity within the cluster
        return np.mean(cosine_similarity(cluster_emb))

    # Report the number of items in each cluster
    cluster_counts = df.groupby("cluster").size()
    for cluster_id, count in cluster_counts.items():
        info(f"Cluster {cluster_id}: {count} items")

    # Group by cluster and apply the function in parallel
    info("Computing average similarity scores for each cluster ...")
    cluster_scores = df.groupby("cluster").apply(compute_cluster_avg_similarity)
    avg_sim_scores = cluster_scores.dropna().to_dict()
    return avg_sim_scores


def _run_hdbscan_assign(
    df: pandas.DataFrame,
    alpha: float,
    cluster_selection_epsilon: float,
    cluster_selection_method: str,
    algorithm: str,
    min_cluster_size: int,
    min_samples: int,
    use_pca: bool,
    use_cuhdbscan: bool,
    batch: int,
    max_batch: int,
    core_dist_n_jobs: int,
) -> pandas.DataFrame:
    """
    Cluster the features using HDBSCAN
    :param alpha:  The alpha parameter for HDBSCAN
    :param cluster_selection_epsilon:  The epsilon parameter for HDBSCAN
    :param algorithm:  The algorithm to use for clustering, 'best' or 'generic' or 'prims_kdtree' or 'boruvka_kdtree'
    :param cluster_selection_method:  The method to use for cluster selection, 'leaf' or 'eom'
    :param min_cluster_size:  The minimum number of samples in a cluster
    :param min_samples:   The number of samples in a neighborhood for a point
    :param use_pca:  Whether to use PCA for dimensionality reduction
    :param use_cuhdbscan:  Whether to use cuHDBSCAN for clustering; requires a GPU with CUDA support
    :param batch:  Batch number for logging
    :param max_batch:  Maximum batch number for logging
    :return: pandas.DataFrame with the cluster assignments
    """
    info(
        f"Clustering using HDBSCAN with: "
        f"sdcat version {sdcat_version},"
        f"core_dist_n_jobs: {core_dist_n_jobs},"
        f"batch {batch} of {max_batch},"
        f"alpha {alpha},"
        f"algorithm {algorithm},"
        f"cluster_selection_epsilon {cluster_selection_epsilon},"
        f"min_samples {min_samples},"
        f"min_cluster_size {min_cluster_size},"
        f"cluster_selection_method {cluster_selection_method},"
        f"use_pca {use_pca},"
        f"use_cuhdbscan {use_cuhdbscan}"
    )
    if use_cuhdbscan and not have_gpu:
        warn("cuHDBSCAN is requested but no GPU is available. Falling back to CPU HDBSCAN.")
        use_cuhdbscan = False

    # Get the number of samples which is the number of rows in the dataframe
    num_samples = df.shape[0]

    info(f"Stacking {len(df)} features ...")
    features = np.stack(df["embedding"].values)
    # Add in other features if present
    if df.shape[1] > 1:
        for col in df.columns:
            if col != "embedding":
                features = np.concatenate((features, df[col].values.reshape(-1, 1)), axis=1)

    if use_pca:
        info("Reducing to 100 dimensions using PCA...")
        pca = PCA(n_components=100, random_state=42)
        x = pca.fit_transform(features)
    else:
        x = features

    # Cluster the features using HDBSCAN
    if have_gpu and use_cuhdbscan:
        info(f"Running HDBSCAN on {num_samples} samples for batch {batch} of {max_batch} on GPU...")
        # L2 norm the data for cuHDBSCAN; used together with Euclidean distance as an approximation of cosine distance
        x_norm = normalize(x, norm="l2")
        scan = cuHDBSCAN(
            prediction_data=True,
            metric="euclidean",
            allow_single_cluster=True,
            algorithm=algorithm,
            min_cluster_size=min_cluster_size,
            min_samples=min_samples,
            alpha=alpha,
            cluster_selection_epsilon=cluster_selection_epsilon,
            cluster_selection_method=cluster_selection_method,
        )
        labels = scan.fit_predict(x_norm)
    else:
        info(
            f"Running HDBSCAN on {num_samples} samples for batch {batch} of {max_batch} on CPU core_dist_n_jobs {core_dist_n_jobs}..."
        )
        # Compute the cosine similarity matrix
        cosine_sim_matrix = cosine_similarity(x)
        distance_matrix = 1 - cosine_sim_matrix
        x = distance_matrix.astype(np.float64)
        scan = HDBSCAN(
            core_dist_n_jobs=core_dist_n_jobs,
            prediction_data=True,
            metric="precomputed",
            algorithm=algorithm,
            allow_single_cluster=True,
            min_cluster_size=min_cluster_size,
            min_samples=min_samples,
            alpha=alpha,
            cluster_selection_epsilon=cluster_selection_epsilon,
            cluster_selection_method=cluster_selection_method,
        )
        labels = scan.fit_predict(x)

    info(f"Number of clusters including unassigned -1 cluster: {len(np.unique(labels))}")

    df_ = df.copy()
    # Save the probabilities and cluster_batch for each cluster which is used for merging
    df_["HDBSCAN_probability"] = scan.probabilities_
    df_["cluster"] = labels
    # Set the cluster batch based on the batch number and cluster id,
    # Create a UUID for each cluster batch to ensure uniqueness
    # e.g. "9476e0d0-3ee3-492e-816c-050f02e07bdb_00000002" for batch 0 and cluster label 2
    # or "-1" for noise
    batch_uuid = str(uuid.uuid4())

    def set_cluster_batch(x):
        if x != -1:
            return f"{batch:08d}{x:08d}_{batch_uuid}"
        return "-1"

    df_["cluster_batch"] = df_["cluster"].apply(set_cluster_batch)
    unique_clusters = df_["cluster"].unique()
    info(f"Batch {batch} index: {df_.index[0]} to {df_.index[-1]}")
    info(
        f"Maximum cluster id: {labels.max()} minimum cluster id: {labels.min()} unique clusters: {len(unique_clusters)} noise: {np.sum(labels == -1)}"
    )
    return df_


def cluster_vits(
    prefix: str,
    model: str,
    df_dets: pd.DataFrame,
    output_path: Path,
    alpha: float,
    cluster_selection_epsilon: float,
    cluster_selection_method: str,
    algorithm: str,
    min_similarity: float,
    min_cluster_size: int,
    min_samples: int,
    device: str = "cpu",
    use_vits: bool = False,
    use_pca: bool = False,
    use_cuhdbscan=False,
    skip_visualization: bool = False,
    remove_bad_images: bool = False,
    roi: bool = False,
    vits_batch_size: int = 32,
    hdbscan_batch_size: int = 50000,
    allowable_classes: list = None,
) -> dict or None:
    """Cluster the crops using the VITS embeddings.
    :param prefix:  A unique prefix to save artifacts from clustering
    :param model: The model to use for clustering
    :param df_dets: The dataframe with the detections
    :param output_path: The output path to save the clustering artifacts to
    :param roi:  Whether the detections are already cropped to the ROI
    :param cluster_selection_epsilon: The epsilon parameter for HDBSCAN
    :param cluster_selection_method: The method to use for cluster selection, 'leaf' or 'eom'
    :param algorithm:  The algorithm to use for clustering, 'best' or 'generic' or 'prims_kdtree' or 'boruvka_kdtree'
    :param alpha: The alpha parameter for HDBSCAN
    :param min_similarity: The minimum similarity score to use for -1 cluster reassignment
    :param min_cluster_size: The minimum number of samples in a cluster
    :param min_samples:The number of samples in a neighborhood for a point
    :param device: The device to use for clustering, 'cpu' or 'cuda' or 'cuda:0' or 'cuda:1'
    :param use_vits: Set to using the predictions from the vits cluster model
    :param skip_visualization: Whether to skip the visualization of the clusters
    :param remove_bad_images: Whether to remove bad images from the clustering
    :param use_pca: Whether to use PCA for dimensionality reduction
    :param use_cuhdbscan: Whether to use cuHDBSCAN for clustering; requires a GPU with CUDA support
    :param vits_batch_size: The batch size to use for embedding extraction; maximize for speed for your GPU
    :param hdbscan_batch_size: The batch size to use for clustering with HDBSCAN; maximize for speed for your GPU
    :param allowable_classes: A list of classes to allow in the clustering; if None, all classes are allowed
    :return:  a dictionary with a summary."""
    warnings.filterwarnings("ignore")
    # If there are no detections, return an empty dataframe
    if df_dets.empty:
        warn("No detections found")
        return None

    # If the detections are not cropped, crop them to a square
    if not roi:
        # Only crop if needed
        existing = df_dets["crop_path"].apply(lambda x: os.path.exists(x))
        num_square = existing.sum()

        if num_square == len(df_dets):
            info(f"All {len(df_dets)} detections are already cropped to square")
        else:
            info(f"Cropping {len(df_dets)} detections...")
            # Filter rows that need cropping
            rows_to_crop = df_dets[~existing]._to_pandas()

            # Crop by grouped image_path which is more efficient
            grouped = rows_to_crop.groupby("image_path")
            info(f"Cropping {len(rows_to_crop)} detections...")

            with ThreadPoolExecutor(max_workers=os.cpu_count()) as executor:
                futures = [executor.submit(crop_all_square_images, group, df, 224, 0) for group, df in grouped]
                for _ in tqdm(as_completed(futures), total=len(futures)):
                    pass

    # Drop any rows with crop_path that have files that don't exist or are zero - sometimes the crops fail
    num_before = len(df_dets)
    df_dets = df_dets[df_dets["crop_path"].apply(lambda x: os.path.exists(x))]
    df_dets = df_dets[df_dets["crop_path"].apply(lambda x: os.path.getsize(x) > 0)]
    num_after = len(df_dets)
    if num_before != num_after:
        info(f"Dropped {num_before - num_after} detections with missing crop_path files")
    if df_dets.empty:
        warn("No detections found")
        return None

    df_dets = df_dets.sort_index()
    df_dets["exemplar"] = 0
    df_dets["cluster"] = -1  # -1 means noise or unassigned
    df_dets["cluster_batch"] = "-1"
    df_dets["HDBSCAN_probability"] = 0

    # Get the list of images to crop
    crop_paths = df_dets["crop_path"].values

    # Count how many files have the .npy extension
    num_cached = sum([has_cached_embedding(model, filename) for filename in crop_paths])
    info(f"Found {num_cached} cached embeddings for {len(crop_paths)} images")

    # Skip the embedding extraction if all the embeddings are cached
    if num_cached != len(crop_paths):
        debug(f"Extracted embeddings from {len(crop_paths)} images using model {model}...")
        compute_norm_embedding(model, crop_paths, device, vits_batch_size)

    def load_model_results(crop_path):
        try:
            _, labels, scores = fetch_embedding(model, crop_path)
            label = labels[0]
            label_s = labels[1]
            score = scores[0]
            score_s = scores[1]
            return pandas.Series({"class": label, "score": score, "class_s": label_s, "score_s": score_s})
        except IndexError:
            return pandas.Series({"class": "Unknown", "score": 0, "class_s": "Unknown", "score_s": 0})

    if use_vits:
        info(f"Loading ViTS model {model} results into dataframe ...")

        # Add in column if missing and apply the function to each row in the dataframe
        results_df = df_dets["crop_path"].apply(load_model_results).apply(pandas.Series)
        df_dets["class"] = results_df["class"]
        df_dets["score"] = results_df["score"]
        df_dets["class_s"] = results_df["class_s"]
        df_dets["score_s"] = results_df["score_s"]

        # If allowable_classes is provided, filter the dataframe to only include those classes and reset the index
        if allowable_classes:
            df_dets = df_dets[df_dets["class"].isin(allowable_classes)].reset_index(drop=True)

            if df_dets.empty:
                print(f"No detections left after filtering by allowable classes {allowable_classes}")
                return None

    if not (output_path / prefix).exists():
        (output_path / prefix).mkdir(parents=True)

    # Remove everything except ancillary data to include in clustering
    columns = [
        "x",
        "y",
        "xx",
        "xy",
        "w",
        "h",
        "image_width",
        "image_height",
        "cluster_id",
        "cluster",
        "score",
        "class",
        "score_s",
        "class_s",
        "image_path",
        "crop_path",
    ]
    # Check if the columns exist in the dataframe
    if all(col in df_dets.columns for col in columns):
        ancillary_df = df_dets.drop(
            columns=[
                "x",
                "y",
                "xx",
                "xy",
                "w",
                "h",
                "image_width",
                "image_height",
                "cluster_id",
                "cluster",
                "score",
                "cluster_s",
                "score_s",
                "class",
                "image_path",
                "crop_path",
            ]
        )
    else:
        ancillary_df = None

    if ancillary_df:
        # Add in any numerical ancillary data and replace NaNs with 0
        numerical_df = ancillary_df.select_dtypes(include=["float", "int"])
        numerical_df = numerical_df.fillna(0)
        # Normalize the numerical data from 0 to 1 - this is an important step!
        ancillary_df = (numerical_df - numerical_df.min()) / (numerical_df.max() - numerical_df.min())

    # Cluster
    num_batches = int(np.ceil(len(crop_paths) / hdbscan_batch_size))

    info(f"Remove any existing cluster grid images in the output_path in {(output_path / prefix)}")
    cluster_grids = (output_path / prefix).rglob(f"{prefix}_*cluster*.png")
    for c in cluster_grids:
        c.unlink()

    info(f"Processing images in batches of {hdbscan_batch_size} ...")

    def process_batch(i, n_jobs, df, df_ancillary=None):
        """
        Process a batch of images and return the clustered dataframe.
        Disable df_ancillary for now
        """
        start = df.index[0]
        end = df.index[-1]
        info(f"Processing batch {i + 1} of {num_batches} {start} to {end}...")
        filepaths = df["crop_path"].values.tolist()
        info(f"Fetching batch {i + 1}  embeddings for {len(filepaths)} images...")
        df["embedding"] = [fetch_embedding(model, filename)[0] for filename in filepaths]
        info(f"Done fetching batch {i + 1} embeddings for {len(filepaths)} images")

        # Drop any rows where the embedding is None or empty
        df = df[df["embedding"].apply(lambda x: x is not None and len(x) > 0)]

        if remove_bad_images:
            info(f"Cleaning bad images from {len(df)} images")
            bad_images = clean_bad_images(filepaths)
            bad_indexes = df[df["crop_path"].isin(bad_images)].index.tolist()
            info(
                f"Removed {len(bad_images)} of {len(filepaths)} images using cleanvision in batch {i + 1} of {num_batches}..."
            )
            df = df.drop(index=bad_indexes)
            # All detections in this batch are bad, so skip clustering
            if df.empty:
                warn(f"No detections left in batch {i + 1} of {num_batches} after cleaning bad images")
                return None

        # Drop unnecessary columns
        keep_columns = ["area", "saliency", "w", "h", "embedding", "crop_path"]
        df = df[[col for col in df.columns if col in keep_columns]]

        # Merge ancillary data
        if df_ancillary is not None:
            for filename in crop_paths[start:end]:
                ancillary_data_row = df_ancillary.loc[df_ancillary["crop_path"] == filename]
                if ancillary_data_row.empty:
                    ancillary_data = pd.Series([0] * len(df_ancillary.columns), index=df_ancillary.columns)
                else:
                    ancillary_data = ancillary_data_row.select_dtypes(include=["float", "int"]).iloc[0]
                df.loc[df_ancillary["crop_path"] == filename, ancillary_data.index] = ancillary_data

        # Drop the crop_path column if it exists as it is not numerical and not needed for clustering
        df = df.drop(columns=["crop_path"], errors="ignore")

        # Run clustering
        return _run_hdbscan_assign(
            df=df,
            alpha=alpha,
            cluster_selection_epsilon=cluster_selection_epsilon,
            cluster_selection_method=cluster_selection_method,
            algorithm=algorithm,
            min_cluster_size=min_cluster_size,
            min_samples=min_samples,
            use_pca=use_pca,
            batch=i,
            max_batch=num_batches,
            core_dist_n_jobs=n_jobs,
            use_cuhdbscan=use_cuhdbscan,
        )

    # Calculate the number of batches
    num_batches = int(np.ceil(len(df_dets) / hdbscan_batch_size))

    # If the number of batches is less than 2, set the core_dist_n_jobs to the number of CPU cores assuming the core
    # can handle the load of two batches at a time
    if num_batches > 1:
        core_dist_n_jobs = 1
        info(f"Using {core_dist_n_jobs} CPUs for HDBSCAN")
    else:
        # Set the number of processes to the number of batches
        core_dist_n_jobs = int(os.cpu_count())
        info(f"Using {core_dist_n_jobs} CPUs for HDBSCAN")

    # Run batches in parallel
    with ThreadPoolExecutor(max_workers=core_dist_n_jobs) as executor:
        futures = [
            executor.submit(
                process_batch,
                i,
                core_dist_n_jobs,
                df_dets.iloc[i * hdbscan_batch_size : min((i + 1) * hdbscan_batch_size, len(crop_paths))]._to_pandas(),
            )
            for i in range(num_batches)
        ]
        # Wait for all the futures to complete
        for result in tqdm(as_completed(futures), total=len(futures)):
            if result:
                df_batch = result.result()
                info(f"Batch {df_batch.index[0]} to {df_batch.index[-1]} completed")
                if df_batch is not None:
                    if df_batch.empty:
                        warn(f"Batch {df_batch.index[0]} to {df_batch.index[-1]} is empty after clustering")
                    else:
                        df_dets.update(df_batch["cluster_batch"])
                        df_dets.update(df_batch["cluster"])
                        df_dets.update(df_batch["HDBSCAN_probability"])

    info("All batch clusters processed")
    if df_dets.empty:
        warn("No detections left after clustering")
        return None

    # Drop any rows with NaN values
    df_dets = df_dets.dropna()
    if df_dets.empty:
        warn("No detections left after dropping NaN values")
        return None

    # Create an increasing array of integers based on the unique cluster_batch values, except for the noise
    info("Mapping clusters to unique integers ...")
    non_noise_df = df_dets[df_dets["cluster"] != -1]

    if non_noise_df.empty:
        warn("No clusters found")
        return None

    unique_cluster_batch = non_noise_df.groupby("cluster_batch")
    cluster_batch_group = {cg: i for i, cg in enumerate(unique_cluster_batch.groups)}
    cluster_batch_group["-1"] = -1  # Ensure noise cluster is mapped to -1

    info(f"Mapping {len(cluster_batch_group)} unique cluster batches to integers ...")
    df_dets["cluster_org"] = df_dets["cluster"]  # Keep the original cluster for reference
    values = df_dets["cluster_batch"].values
    df_dets["cluster"] = [cluster_batch_group.get(c, -1) for c in values]  # Map to contiguous integers

    unique_clusters = df_dets["cluster"].unique()
    info(f"Index: {df_dets.index[0]} to {df_dets.index[-1]}")
    info(
        f"Remapped maximum cluster id: {df_dets['cluster'].values.max()} minimum cluster id: {df_dets['cluster'].values.min()} unique clusters: {len(unique_clusters)}"
    )

    # Compute the exemplars for each cluster
    exemplar_emb, cluster = _compute_exemplars(df_dets, model, device, vits_batch_size)

    # Reassign noise (-1) cluster to the nearest exemplar cluster based on cosine similarity
    df_dets_re = _reassign_noise(df_dets, exemplar_emb, cluster, min_similarity, model)

    # Merge by similarity to the exemplar embeddings
    df_dets_final = _similarity_merge(df_dets_re, exemplar_emb, cluster, min_similarity, model)

    # We may have fewer clusters after merging, so reassign the cluster ids to contiguous integers
    non_noise_df = df_dets_final[df_dets_final["cluster"] != -1]
    unique_cluster = non_noise_df.groupby("cluster")
    cluster_group = {cg: i for i, cg in enumerate(unique_cluster.groups)}
    cluster_group["-1"] = -1  # Ensure noise cluster is mapped to -1

    info(f"Mapping {len(unique_clusters)} unique cluster to contiguous integers ...")
    values = df_dets_final["cluster"].values
    df_dets["cluster"] = [cluster_group.get(c, -1) for c in values]  # Map to contiguous integers

    # Get the average similarity within and across all clusters
    avg_sim_scores = _compute_avg_sim(df_dets_final, model)
    avg_similarity = np.mean(list(avg_sim_scores.values()))

    info(f"Average similarity: {avg_similarity:.2f} min {min_similarity:.2f}  ")
    unique_clusters = df_dets_final["cluster"].unique()

    # Drop the -1 value which is the noise cluster
    if -1 in unique_clusters:
        unique_clusters = unique_clusters[unique_clusters != -1]
    unique_clusters.sort()

    if len(unique_clusters) == 0:
        warn("No clusters found")

    info(f"Found {len(unique_clusters)} clusters with an average similarity of {avg_similarity:.2f} ")

    num_samples = df_dets.shape[0]
    clustered = df_dets["cluster"].values != -1
    coverage = np.sum(clustered) / num_samples
    info(f"Coverage: {coverage:.2f} ({np.sum(clustered)}/{num_samples})")

    hdbscan_params = {
        "min_cluster_size": min_cluster_size,
        "min_samples": min_samples,
        "cluster_selection_method": "leaf",
        "metric": "precomputed",
        "algorithm": algorithm,
        "alpha": alpha,
        "cluster_selection_epsilon": cluster_selection_epsilon,
        "use_pca": use_pca,
    }
    if use_cuhdbscan and have_gpu:
        hdbscan_params["use_cuhdbscan"] = True
        hdbscan_params["metric"] = "l2"

    df_dets_final.to_csv(output_path / f"{prefix}_cluster_detections.csv")
    info(f'Saved {output_path / f"{prefix}_cluster_detections.csv"}')
    df_dets_final.to_parquet(output_path / f"{prefix}_cluster_detections.parquet")
    info(f'Saved {output_path / f"{prefix}_cluster_detections.parquet"}')

    # Return a summary of the clusters
    return _summarize_clusters(df_dets_final, output_path, prefix, avg_sim_scores, model, skip_visualization, hdbscan_params)



# sdcat, Apache-2.0 license
# Filename: sdcat/ml/saliency.py
# Description:  Miscellaneous saliency functions for detecting targets in images using saliency maps
import os
import tempfile
import cv2
import numpy as np
import pandas as pd

from sdcat.cluster.utils import rescale
from sdcat.reflectance.cleaner import specularity_removal
from sdcat.logger import debug, info
from pathlib import Path

# This is a global variable to control whether to save the algorithm results; useful for debugging or presentation
save = True


def compute_saliency(contour, min_std: float, img_saliency: np.ndarray, img_luminance: np.ndarray) -> int:
    """
    Calculate the saliency cost of a contour. Lower saliency contours are more likely to be reflections.
    :param contour: the contour
    :param min_std: minimum standard deviation of the contour to be considered
    :param img_saliency: saliency image
    :param img_luminance: luminance image
    :return: saliency cost (int)
    """
    # Calculate brightness and other statistics of the contour
    mask = np.zeros_like(img_saliency)
    cv2.drawContours(mask, [contour], 0, 255, thickness=cv2.FILLED)
    mean_intensity_l = cv2.mean(img_luminance, mask=mask)[0]
    mean_intensity_s = cv2.mean(img_saliency, mask=mask)[0]

    # Calculate area of the contour
    area = cv2.contourArea(contour)

    # Calculate variance of pixel intensities within the contour
    x, y, w, h = cv2.boundingRect(contour)
    roi = img_luminance[y : y + h, x : x + w]
    variance = np.var(roi)
    std = np.std(roi)

    # Create a factor to normalize the cost for different image sizes
    factor = img_luminance.size / 10000

    # The cost function penalizes smaller areas with low variance
    cost = (area * (mean_intensity_l + mean_intensity_s + 0.1 * area) - variance) / factor

    # If the std is too low, then the contour is not interesting; set the cost to 1
    if std < min_std:
        cost = 1

    return int(cost)


def process_contour(contours: np.ndarray, min_std: float, img_s: np.ndarray, img_l: np.ndarray) -> np.ndarray:
    """
    Process a single contour. Save the results to a csv file.
    This is a separate function, so it can be run in parallel.
    :param contours:  List of contours
    :param min_std: minimum standard deviation of the contour to be considered
    :param img_s: Saliency image
    :param img_l: Luminance image
    :return: dataframe of blobs
    """
    df = pd.DataFrame()
    for c in contours:
        x, y, w, h = cv2.boundingRect(c)

        area = cv2.contourArea(c)
        saliency = compute_saliency(c, min_std, img_s, img_l)

        debug(f"Found blob area: {area}, saliency: {saliency}, area: {area}")
        df = pd.concat(
            [
                df,
                pd.DataFrame(
                    {
                        "image_path": "",
                        "class": "Unknown",
                        "score": 0.0,
                        "area": area,
                        "saliency": saliency,
                        "x": x,
                        "y": y,
                        "xx": x + w,
                        "xy": y + h,
                        "w": w,
                        "h": h,
                    },
                    index=[0],
                ),
            ]
        )

    return df


def smooth_reflections(image: np.ndarray, show=False) -> np.ndarray:
    """
    Smooth reflections by filling in bright blobs with the mean outside the blob followed
    by gaussian smoothing
    :param image:
    :return:
    """
    # Threshold the image to create a binary mask of pixels > 180
    _, mask = cv2.threshold(image, 180, 255, cv2.THRESH_BINARY + cv2.THRESH_OTSU)

    # Find contours in the binary mask
    contours, _ = cv2.findContours(mask, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_SIMPLE)

    filtered_contours = [cnt for cnt in contours if cv2.contourArea(cnt) < 10000]

    # Fill in all contours with 128
    cv2.drawContours(image, filtered_contours, -1, 128, thickness=cv2.FILLED)

    image = cv2.GaussianBlur(image, (15, 15), 0)
    if show:
        cv2.imshow("Smooth reflections", image)
        if save:
            cv2.imwrite("my_smooth_reflections.jpg", image)
        cv2.waitKey(0)
    return image


def extract_blobs(
    min_std: float, block_size: int, saliency_map: np.ndarray, img_color: np.ndarray, show=False
) -> (pd.DataFrame, np.ndarray):
    """
    Extract blobs from a saliency map
    :param min_std: minimum standard deviation of the contour to be considered
    :param block_size: block size for adaptive thresholding
    :param saliency_map: normalized saliency map
    :param img_color: color image used when showing the results
    :param show: True to show the results
    :return: pandas dataframe of blobs, image with contours drawn
    """

    # Blur the saliency map to remove noise
    saliency_map = cv2.GaussianBlur(saliency_map, (15, 15), 0)

    if show:
        # Display the thresholded saliency map
        cv2.imshow("Gaussian Blur Saliency Map", saliency_map.astype(np.uint8))
        if save:
            cv2.imwrite("my_gaussian_blur_saliency_map.jpg", saliency_map.astype(np.uint8))
        cv2.waitKey(0)

    # Threshold the saliency map using gradient thresholding
    saliency_map_thres_c = cv2.adaptiveThreshold(
        saliency_map.astype(np.uint8),
        255,  # Max pixel value
        cv2.ADAPTIVE_THRESH_GAUSSIAN_C,
        cv2.THRESH_BINARY,
        block_size,  # Block size (size of the local neighborhood)
        3,  # Constant subtracted from the mean
    )

    # Invert the saliency map so that the salient regions are white
    saliency_map_thres_c = cv2.bitwise_not(saliency_map_thres_c)

    # Connect blobs by dilating then eroding them
    # For images > 3000 pixels, use a 9x9 kernel, otherwise use a 5x5 kernel
    width, height = saliency_map_thres_c.shape
    info(f"Saliency map image size: {width} x {height}")
    if width > 3000 or height > 3000:
        kernel1 = np.ones((5, 5), np.uint8)
        kernel2 = np.ones((3, 3), np.uint8)
        saliency_map_thres_c = cv2.dilate(saliency_map_thres_c, kernel1, iterations=4)
        saliency_map_thres_c = cv2.erode(saliency_map_thres_c, kernel2, iterations=4)
    else:
        kernel1 = np.ones((4, 4), np.uint8)
        kernel2 = np.ones((3, 3), np.uint8)
        saliency_map_thres_c = cv2.dilate(saliency_map_thres_c, kernel1, iterations=1)
        saliency_map_thres_c = cv2.erode(saliency_map_thres_c, kernel2, iterations=1)

    if show:
        # Display the thresholded saliency map
        cv2.imshow("Threshold Saliency Map", saliency_map_thres_c.astype(np.uint8))
        if save:
            cv2.imwrite("my_thresholded_saliency_map.jpg", saliency_map_thres_c.astype(np.uint8))
        cv2.waitKey(0)

    contours, _ = cv2.findContours(saliency_map_thres_c, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_TC89_L1)

    # Draw the contours on the image with a purple line border thickness of 2 px
    contour_img = img_color.copy()
    cv2.drawContours(contour_img, contours, -1, (81, 12, 51), 2)

    if show:
        cv2.imshow(f"{len(contours)} Contours", contour_img)
        if save:
            cv2.imwrite("my_contours.jpg", contour_img)
        cv2.waitKey(0)

    # Get the luminance to use in saliency calculation
    img_lum = cv2.cvtColor(img_color, cv2.COLOR_BGR2LAB)[:, :, 0]

    info(f"Found {len(contours)} contours. Processing contours...")
    df = process_contour(contours, min_std, saliency_map_thres_c.astype(np.uint8), img_lum)

    if show:
        for i, row in df.iterrows():
            x = row["x"]
            y = row["y"]
            w = row["w"]
            h = row["h"]
            saliency = row["saliency"]
            cv2.rectangle(contour_img, (x, y), (x + w, y + h), (247, 108, 0), 2)
            saliency = np.round(saliency)
            cv2.putText(contour_img, f"{saliency}", (x, y), cv2.FONT_HERSHEY_SIMPLEX, 0.5, (255, 255, 255), 2)

        cv2.imshow(f"{len(df)} Bounding Boxes around Salient Regions", contour_img)
        if save:
            cv2.imwrite("my_bounding_boxes.jpg", contour_img)
        cv2.waitKey(0)
        cv2.destroyAllWindows()

    return df, contour_img


def fine_grained_saliency(image, clahe=False, show=False) -> np.ndarray:
    """
    Compute a fine-grained saliency map and normalize it
    :param image:  image
    :param clahe: True to run CLAHE on the luminance channel
    :param show:  True to show the results
    :return: normalized saliency map
    """

    img_hsv = cv2.cvtColor(image, cv2.COLOR_BGR2HSV)
    img_lab = cv2.cvtColor(image, cv2.COLOR_BGR2LAB)

    # Blur the saturation channel to remove noise
    img_hsv[:, :, 1] = cv2.GaussianBlur(img_hsv[:, :, 1], (15, 15), 0)

    # Run CLAHE on the image luminance channel
    if clahe:
        clahe = cv2.createCLAHE(clipLimit=4.0, tileGridSize=(6, 6))
        filtered_lum = clahe.apply(img_lab[:, :, 0])
    else:
        filtered_lum = img_lab[:, :, 0]
    filtered_saturation = img_hsv[:, :, 1]

    img_sl = cv2.merge((filtered_saturation, filtered_lum))

    # Initialize an empty saliency map
    saliency_map = np.zeros_like(image[:, :, 0], dtype=np.float32)

    blurred_saturation = cv2.GaussianBlur(img_sl[:, :, 1], (5, 5), 0)
    blurred_lum = cv2.GaussianBlur(img_sl[:, :, 1], (5, 5), 0)

    # Calculate the center and surround regions for each channel
    center_lum = cv2.GaussianBlur(blurred_lum, (3, 3), 2)
    surround_lum = blurred_lum - center_lum

    center_saturation = cv2.GaussianBlur(blurred_saturation, (3, 3), 2)
    surround_saturation = blurred_saturation - center_saturation

    # Normalize the values to the range [0, 255]
    center_lum = cv2.normalize(center_lum, None, 0, 255, cv2.NORM_MINMAX)
    surround_lum = cv2.normalize(surround_lum, None, 0, 255, cv2.NORM_MINMAX)

    center_saturation = cv2.normalize(center_saturation, None, 0, 255, cv2.NORM_MINMAX)
    surround_saturation = cv2.normalize(surround_saturation, None, 0, 255, cv2.NORM_MINMAX)

    # Combine the center and surround regions for each channel
    # Reduce the weight of the saturation channel
    saliency_lum = 1.2 * (center_lum - surround_lum)
    saliency_saturation = (center_saturation - surround_saturation) / 4

    # Combine the saliency maps into the final saliency map
    saliency_level = (saliency_saturation + saliency_lum) / 2

    # Resize the saliency map to the original image size
    saliency_level = cv2.resize(saliency_level, (image.shape[1], image.shape[0]))

    # Accumulate the saliency map at each level
    saliency_map += saliency_level

    # Normalize the final saliency map
    saliency_map = cv2.normalize(saliency_map, None, 0, 255, cv2.NORM_MINMAX)

    # Blur the saliency map to remove noise
    saliency_map = cv2.GaussianBlur(saliency_map, (15, 15), 0)

    if show:
        # Display the original image and the saliency map
        cv2.imshow("Original Image", image)
        cv2.imshow("Fine-Grained Saliency Map", saliency_map.astype(np.uint8))
        if save:
            cv2.imwrite("my_fine_grained_saliency_map.jpg", saliency_map.astype(np.uint8))
            cv2.imwrite("my_original_image.jpg", image)

        cv2.waitKey(0)

    return saliency_map


def run_saliency_detect_bulk(
    spec_removal: bool,
    scale_percent: int,
    images: list,
    out_path: Path,
    min_std: float = 8.0,
    block_size: int = 23,
    clahe: bool = False,
    show: bool = False,
):
    info(f"Processing {len(images)} images")
    for f in images:
        out_csv_file = out_path / f"{f.stem}.csv"
        out_image_file = out_path / f"{f.stem}.jpg" if show else None
        run_saliency_detect(
            spec_removal,
            scale_percent,
            f.as_posix(),
            out_csv_file.as_posix(),
            out_image_file.as_posix() if out_image_file else None,
            min_std,
            block_size,
            clahe,
            show,
        )


def run_saliency_detect(
    spec_removal: bool,
    scale_percent: int,
    in_image_file: str,
    out_csv_file: str,
    out_image_file: str = None,
    min_std: float = 8.0,
    block_size: int = 23,
    clahe: bool = False,
    show: bool = False,
):
    """
    Run the detection algorithm on an saliency map image
    :param spec_removal: True to remove specular highlights
    :param scale_percent: Scale percentage to rescale the image before processing
    :param in_image_file: Input image filename
    :param clahe: True to run CLAHE on the luminance channel
    :param show: True to show the results
    :param out_csv_file: The csv path to save the results
    :param out_image_file: (optional) The image filename to save the results
    :param min_std: minimum standard deviation of the contour to be considered
    :param block_size: block size for adaptive thresholding
    :return:
    """
    info(f"Processing {in_image_file}")
    image_path = Path(in_image_file)
    img_color = cv2.imread(image_path.as_posix())
    img_color_rescaled = rescale(img_color, scale_percent=scale_percent)

    if show:
        cv2.imshow("Original Rescaled", img_color_rescaled)
        if save:
            cv2.imwrite("my_original_rescaled.jpg", img_color_rescaled)
        cv2.waitKey(0)

    # Calculate the scale factor to rescale the detections back to the original image size
    scale_width, scale_height = (
        img_color.shape[0] / img_color_rescaled.shape[0],
        img_color.shape[1] / img_color_rescaled.shape[1],
    )

    if spec_removal:
        # Remove reflections
        img_spc = specularity_removal(img_color_rescaled)
        if show:
            cv2.imshow("Clean Rescaled", img_spc)
            if save:
                cv2.imwrite("my_clean_rescaled.jpg", img_spc)
            cv2.waitKey(0)
    else:
        img_spc = img_color_rescaled

    # Run the fine-grained saliency algorithm
    saliency_map = fine_grained_saliency(img_spc, clahe=clahe, show=show)

    # Extract blobs from the saliency map
    df, contour_img = extract_blobs(min_std, block_size, saliency_map, img_spc, show=show)

    info(f"Found {len(df)} blobs in {image_path}")

    if len(df) == 0:
        return

    if show or out_image_file:
        # Plot boxes on the input frame
        pred_list = df[["x", "y", "xx", "xy", "score", "saliency"]].values.tolist()
        for p in pred_list:
            # Draw a rectangle with yellow line borders of thickness of 3 px
            cv2.rectangle(img_spc, (int(p[0]), int(p[1])), (int(p[2]), int(p[3])), (247, 108, 0), 3)
            # Add saliency score
            cv2.putText(img_spc, f"{int(p[5])}", (int(p[0]), int(p[1])), cv2.FONT_HERSHEY_SIMPLEX, 1.0, (255, 255, 255), 2)

        if show:
            cv2.imshow(f"Found {len(pred_list)} detections", img_spc)
            if save:
                cv2.imwrite("my_detections.jpg", img_spc)
            cv2.waitKey(0)
            cv2.destroyAllWindows()

        if out_image_file:
            # Save the image with the contours drawn
            cv2.imwrite(out_image_file, img_spc)

    # Scale the detections back to the original image size
    df[["x", "xx", "w"]] *= scale_width
    df[["y", "xy", "h"]] *= scale_height
    df["image_path"] = image_path.as_posix()

    # Save the results to a csv file for later processing
    df.to_csv(out_csv_file, mode="w", header=True, index=False)


if __name__ == "__main__":
    # Read all images from the tests/data/kelpflow directory
    # Get the path of this file
    # test_path = Path(__file__).parent.parent.parent / 'tests' / 'data' / 'flirleft'
    # test_path = Path(__file__).parent.parent.parent / 'tests' / 'data' / 'isiis_single'
    # test_path = Path(__file__).parent.parent.parent / 'tests' / 'data' / 'bird'
    # test_path = Path(__file__).parent.parent.parent / 'tests' / 'data' / 'other'
    # test_path = Path(__file__).parent.parent.parent / 'tests' / 'data' / 'whale'
    # test_path = Path(__file__).parent.parent.parent / 'tests' / 'data' / 'front'
    # test_path = Path(__file__).parent.parent.parent / 'tests' / 'data' / 'kelpflow'
    # test_path = Path( __file__).parent.parent.parent / 'tests' / 'data' / 'glare'
    # test_path = Path(__file__).parent.parent.parent / 'tests' / 'data' / 'dolphin'
    # test_path = Path(__file__).parent.parent.parent / 'tests' / 'data' / 'all'
    # test_path = Path(__file__).parent.parent.parent / 'tests' / 'data' / 'otter'
    test_path = Path(__file__).parent.parent.parent / "tests" / "data" / "jelly"
    scale = 50  # Run at 1/2 scale
    out_path = Path(__file__).parent.parent.parent / "tests" / "data" / "out" / "jelly"
    out_path.mkdir(parents=True, exist_ok=True)

    # Count the images in the directory
    ext_glob = "*.JPG"
    num_images = len(list(test_path.glob(ext_glob)))

    # Do the work in a temporary directory
    tmp_dir = os.getenv("TMPDIR", tempfile.gettempdir())
    with tempfile.TemporaryDirectory(dir=tmp_dir) as temp_path:
        temp_path = Path(temp_path)

        for in_image_file in test_path.glob(ext_glob):
            run_saliency_detect(
                False,
                scale,
                in_image_file,
                (temp_path / f"{in_image_file.stem}.csv").as_posix(),
                (temp_path / f"{in_image_file.stem}.jpg").as_posix(),
                min_std=8.0,
                block_size=23,
                show=True,
            )

        # Copy the results to the output directory
        for csv in temp_path.glob("*.csv"):
            csv.rename(out_path / csv.name)
        for image in temp_path.glob("*.jpg"):
            image.rename(out_path / image.name)

    info(f"Processed {num_images} images. Results are in {out_path}")



# sdcat, Apache-2.0 license
# Filename: sdcat/cluster/utils.py
# Description: Miscellaneous utility functions for cropping, clustering, and saving detections
import cv2
import numpy as np
import modin.pandas as pd
from typing import List

import pandas
from PIL import Image
from cleanvision import Imagelab
from matplotlib import pyplot as plt
from mpl_toolkits.axes_grid1 import ImageGrid
from pathlib import Path
from sdcat.logger import info, debug, warn, exception
from sdcat import __version__ as sdcat_version


def cluster_grid(
    prefix: str, cluster_sim: float, cluster_id: int, nb_images_display: int, images: List[str], output_path: Path
):
    """
    Cluster visualization; create a grid of images
    :
    :param cluster_sim: Cluster similarity
    :param cluster_size: Size of the cluster
    :param cluster_id: Cluster ID
    :param nb_images_display : Number of images to display in the grid
    :param images: list of images
    :param output_path: output path to save the visualization to
    :return:
    """
    cluster_size = len(images)
    debug(f"Cluster number {cluster_id} size {cluster_size} similarity {cluster_sim}")
    try:
        # Plot a grid for each group of images nb_images_display at a time (e.g. 8x8)
        for i in range(0, len(images), nb_images_display * nb_images_display):
            fig = plt.figure(figsize=(10.0, 10.0))
            grid = ImageGrid(
                fig,
                111,  # similar to subplot(111)
                nrows_ncols=(nb_images_display, nb_images_display),
                # creates nb_images_display x nb_images_display grid of axes
                axes_pad=0.025,
                share_all=True,
                cbar_pad=0.025,
            )
            images_display = images[i : i + nb_images_display * nb_images_display]
            page = i // (nb_images_display * nb_images_display)

            # If we have more than 3 pages, then only display the first 3 pages
            # There can be a large number of pages for detections in common classes
            if page > 3:
                break

            total_pages = len(images) // (nb_images_display * nb_images_display)
            for j, image in enumerate(images_display):
                try:
                    image_square = Image.open(image)
                    grid[j].imshow(image_square)
                except Exception as e:
                    exception(f"Error opening {image} {e}")
                    continue

                grid[j].axis("off")
                grid[j].set_xticklabels([])

            if cluster_id == -1:
                cluster_id = "noise"

            # Add a title to the figure
            if total_pages > 1:
                fig.suptitle(
                    f"{prefix}\nsdcat_version {sdcat_version}\nCluster {cluster_id}, Page: {page} of {total_pages}\nSimilarity: {cluster_sim:.2f}, Size: {cluster_size}",
                    fontsize=14,
                )
            else:
                fig.suptitle(
                    f"{prefix}\nsdcat_version {sdcat_version}\nCluster {cluster_id}\nSimilarity: {cluster_sim:.2f}, Size: {cluster_size}",
                    fontsize=14,
                )

            # Set the background color of the grid to white
            fig.set_facecolor("white")

            # Write the figure to a file
            out = output_path / f"{prefix}_cluster_{cluster_id}_p{page}.png"
            debug(f"Writing {out}")
            fig.savefig(out.as_posix())
            plt.close(fig)
    except Exception as e:
        exception(f"Error creating cluster grid {e}")


def clean_bad_images(filepaths: List[str]) -> List[str]:
    """Remove blurry and near duplicate images."""
    imagelab = Imagelab(filepaths=filepaths)

    # Override the default configurations
    config = {
        "visualize_num_images_per_row": 4,
        "report_num_images": 4,
        "report_max_prevalence": 1.0,  # Default is 0.5, but we want to see all issues
        "report_cell_size": (2, 2),
    }
    imagelab._config = config

    # Detect dark and blurry images
    issue_types = {
        "blurry": {"threshold": 0.52},  # Default is 0.5, but we want to be more strict
        "near_duplicates": {"hash_size": 4, "hash_types": ["whash", "phash"]},
    }
    imagelab.find_issues(issue_types)
    # imagelab.report() Disabling report = this is causing a segfault when run on large datasets

    bad_images = []

    # Remove all low information images
    low_info_images = imagelab.issues[imagelab.issues["is_low_information_issue"]].index
    if len(low_info_images) > 0:
        info(f"Removing {len(low_info_images)} low information images")
        bad_images.extend(low_info_images)

    # Remove all dark images
    dark_images = imagelab.issues[imagelab.issues["is_dark_issue"]].index
    if len(dark_images) > 0:
        info(f"Removing {len(dark_images)} dark images")
        bad_images.extend(dark_images)

    # Remove all blurry images
    blurry_images = imagelab.issues[imagelab.issues["is_blurry_issue"]].index
    if len(blurry_images) > 0:
        info(f"Removing {len(blurry_images)} blurry images")
        bad_images.extend(blurry_images)

    # Remove one image from each near duplicate set (keep the best blurry score)
    near_duplicates_image_sets = imagelab.info["near_duplicates"]["sets"]
    near_duplicates_scores = (
        imagelab.issues[imagelab.issues["is_near_duplicates_issue"]]
        .sort_values(by="blurry_score")
        .reset_index()[["index", "blurry_score"]]
        .values.tolist()
    )

    info(f"Removing {len(near_duplicates_image_sets)} duplicated image sets, keeping the least blurry image")
    for dup_set in near_duplicates_image_sets:
        if len(dup_set) <= 1:
            continue
        # Keep the image with the **highest** blurry_score (least blurry)
        best_image = max(
            (entry for entry in near_duplicates_scores if entry[0] in dup_set), key=lambda x: x[1], default=None
        )
        if best_image is None:
            continue
        to_remove = [img for img in dup_set if img != best_image[0]]
        bad_images.extend(to_remove)
    return list(bad_images)


def compute_square_coordinates(row) -> List[int]:
    """
    Compute square coordinates for cropping an image to a square, padding the shorter dimension.

    This also adjusts the crop to make sure the crop is fully in the frame, otherwise the crop that
    exceeds the frame is filled with black bars - these produce clusters of "edge" objects instead
    of the detection
    """
    x1 = int(row.image_width * row.x)
    y1 = int(row.image_height * row.y)
    x2 = int(row.image_width * row.xx)
    y2 = int(row.image_height * row.xy)
    width = x2 - x1
    height = y2 - y1
    shorter_side = min(height, width)
    longer_side = max(height, width)
    delta = abs(longer_side - shorter_side)

    # Divide the difference by 2 to determine how much padding is needed on each side
    padding = delta // 2

    # Add the padding to the shorter side of the image
    if width == shorter_side:
        x1 -= padding
        x2 += padding
    else:
        y1 -= padding
        y2 += padding

    # Make sure that the coordinates don't go outside the image
    # If they do, adjust by the overflow amount
    if y1 < 0:
        y1 = 0
        y2 += abs(y1)
        if y2 > row.image_height:
            y2 = row.image_height
    elif y2 > row.image_height:
        y2 = row.image_height
        y1 -= abs(y2 - row.image_height)
        if y1 < 0:
            y1 = 0
    if x1 < 0:
        x1 = 0
        x2 += abs(x1)
        if x2 > row.image_width:
            x2 = row.image_width
    elif x2 > row.image_width:
        x2 = row.image_width
        x1 -= abs(x2 - row.image_width)
        if x1 < 0:
            x1 = 0
    return [x1, y1, x2, y2]


def crop_rotated_fill(image, cx, cy, w, h, angle) -> np.ndarray:
    """
    Crop a rotated rectangle from an image and fill the background with the mean color of the image.
    """
    # Compute mean color of the image (across all pixels and channels)
    mean_color = image.mean(axis=(0, 1)).astype(np.uint8).tolist()

    # Create rotated rectangle
    rect = ((cx, cy), (w, h), angle)
    box = cv2.boxPoints(rect).astype(np.float32)

    # Destination points for upright box
    dst_pts = np.array([[0, 0], [w - 1, 0], [w - 1, h - 1], [0, h - 1]], dtype=np.float32)

    M = cv2.getPerspectiveTransform(box, dst_pts)

    # Warp with border filled by mean color
    warped = cv2.warpPerspective(image, M, (int(w), int(h)), borderMode=cv2.BORDER_CONSTANT, borderValue=mean_color)
    return warped


def crop_all_square_images(image_path: str, detections: pandas.DataFrame, square_dim: int, rotation: int):
    """
    Crop the image to a square padding the shortest dimension, then resize it to square_dim x square_dim
    Optimized for cropping all detections within an image
    :param image_path: Path to the image
    :param detections: Dataframe with the detections
    :param square_dim: dimension of the square image
    :param rotation: Rotation angle in degrees to apply to the image before cropping
    """
    try:
        if not Path(image_path).exists():
            warn(f"Skipping because the image {image_path} does not exist")
            return

        # If the number of crops that already exist is equal to the number of detections, return early
        detections["exists"] = detections["crop_path"].apply(lambda x: Path(x).exists())
        if detections["exists"].all():
            debug(f"All crops for {image_path} already exist, skipping cropping")
            return

        if rotation > 0:
            # Read the image converting to BGR format for OpenCV
            image = cv2.imread(image_path)
            image_rgb = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)

            # Iterate over the detections and crop each one
            for index, row in detections.iterrows():
                # If the crop already exists, skip it to save time
                if Path(row.crop_path).exists():
                    continue
                x1, y1, x2, y2 = compute_square_coordinates(row)
                img_cropped = crop_rotated_fill(
                    image_rgb, cx=int((x1 + x2) / 2), cy=int((y1 + y2) / 2), w=x2 - x1, h=y2 - y1, angle=int(rotation)
                )

                # Resize the image to square_dim x square_dim
                cropped = cv2.resize(img_cropped, (224, 224), interpolation=cv2.INTER_LANCZOS4)
                # Convert back to BGR format for saving
                cropped = cv2.cvtColor(cropped, cv2.COLOR_RGB2BGR)
                cv2.imwrite(row.crop_path, cropped)
                debug(f"Cropping {image_path} to {row.crop_path} with coordinates ({x1}, {y1}, {x2}, {y2})")
        else:
            img = Image.open(image_path)

            # Iterate over the detections and crop each one
            for index, row in detections.iterrows():
                # If the crop already exists, skip it to save time
                if Path(row.crop_path).exists():
                    continue
                x1, y1, x2, y2 = compute_square_coordinates(row)
                img_cropped = img.crop((x1, y1, x2, y2))
                # Resize the image to square_dim x square_dim
                img_cropped = img_cropped.resize((square_dim, square_dim), Image.LANCZOS)
                debug(f"Cropping {image_path} to {row.crop_path} with coordinates ({x1}, {y1}, {x2}, {y2})")
                img_cropped.save(row.crop_path)
                img_cropped.close()

            img.close()
    except Exception as e:
        exception(f"Error cropping {image_path} {e}")
        raise e


def crop_square_image(row, square_dim: int):
    """
    Crop the image to a square padding the shortest dimension, then resize it to square_dim x square_dim
    :param row:
    :param square_dim: dimension of the square image
    :return:
    """
    try:
        if not Path(row.image_path).exists():
            warn(f"Skipping {row.crop_path} because the image {row.image_path} does not exist")
            return

        if Path(row.crop_path).exists():  # If the crop already exists, skip it
            return

        x1, y1, x2, y2 = compute_square_coordinates(row)

        # Crop the image
        img = Image.open(row.image_path)
        img = img.crop((x1, y1, x2, y2))

        # Resize the image to square_dim x square_dim
        img = img.resize((square_dim, square_dim), Image.LANCZOS)

        # Save the image
        img.save(row.crop_path)
        img.close()

    except Exception as e:
        exception(f"Error cropping {row.image_path} {e}")
        raise e


def rescale(img: np.ndarray, scale_percent: int = 75) -> np.ndarray:
    """
    Rescale an image
    :param img: Image to rescale
    :param scale_percent: Scale percentage
    :return: Rescaled image
    """
    width, height = img.shape[:2]
    width = int(width * scale_percent / 100)
    height = int(height * scale_percent / 100)
    dim = (height, width)

    # Resize the image to the new dimensions exactly
    img_rescaled = cv2.resize(img, dim, interpolation=cv2.INTER_AREA)
    return img_rescaled


def filter_images(min_area: int, max_area: int, min_saliency: int, min_score: float, df: pd.DataFrame) -> pd.DataFrame:
    """
    Filter the dataframe to remove images that are too small or have low saliency
    :param min_area: Minimum area of the image
    :param max_area: Maximum area of the image
    :param min_saliency: Minimum saliency of the image
    :param df: Dataframe to filter
    :return: Filtered dataframe
    """
    # Filter by saliency, area, score or day/night
    size_before = len(df)
    if "saliency" in df.columns:
        df = df[(df["saliency"] > min_saliency) | (df["saliency"] == -1)]
    if "area" in df.columns:
        df = df[(df["area"] > min_area) & (df["area"] < max_area)]
    if "score" in df.columns:
        df = df[(df["score"] > min_score)]
    size_after = len(df)
    info(f"Removed {size_before - size_after} detections outside of area, saliency, or too low scoring")

    return df


def compute_embedding_multi_gpu(model_name: str, images: list, batch_size: int = 32):
    from concurrent.futures import ThreadPoolExecutor
    from sdcat.cluster.embedding import ViTWrapper, compute_embedding_vits
    import torch
    import math

    # Detect all CUDA devices
    devices = [f"cuda:{i}" for i in range(torch.cuda.device_count())]

    # Split the images evenly across GPUs
    def split_batches(images, num_splits):
        batch_size = math.ceil(len(images) / num_splits)
        return [images[i * batch_size : (i + 1) * batch_size] for i in range(num_splits)]

    # Compute embeddings per GPU
    def compute_on_device(device, model_name, images, batch_size):
        vit_wrapper = ViTWrapper(device=device, model_name=model_name)
        compute_embedding_vits(vit_wrapper, images, batch_size)

    def multi_gpu_compute(model_name, images, batch_size):
        image_batches = split_batches(images, len(devices))
        with ThreadPoolExecutor(max_workers=len(devices)) as executor:
            futures = [
                executor.submit(compute_on_device, device, model_name, batch, batch_size)
                for device, batch in zip(devices, image_batches)
            ]
            # Wait for all tasks to complete
            for f in futures:
                f.result()

    multi_gpu_compute(model_name, images, batch_size)


def combine_csv(csv_files: List[Path], temp_path: Path, crop_path: str) -> str:
    from tqdm import tqdm

    output_file = temp_path / "combined.csv"

    info(f"Combining detection files to {output_file}...")
    with open(output_file, "w", encoding="utf-8") as outfile:
        first_file = True
        for file in tqdm(csv_files, desc="Combining detection files", unit="file"):
            # Create a crop directory for each detection file
            crop_root = crop_path / file.stem
            crop_root.mkdir(parents=True, exist_ok=True)
            with open(file, "r", encoding="utf-8") as infile:
                lines = infile.readlines()
                if first_file:
                    header = lines[0].strip() + ",crop_root\n"
                    outfile.writelines(header)  # include header
                    out_text = [line.strip() + f",{crop_root}\n" for line in lines[1:]]
                    outfile.writelines(out_text)
                    first_file = False
                else:
                    out_text = [line.strip() + f",{crop_root}\n" for line in lines[1:]]
                    outfile.writelines(out_text)
    info(f"Combined detection files to {output_file}")
    return str(output_file)



# Credit: Murat Kirtay, The BioRobotics Inst./SSSA/
# Date: 02/11/2016

import cv2
import numpy as np


def derive_graym(impath):
    """The intensity value m is calculated as (r+g+b)/3, yet
    grayscale will do same operation!
    opencv uses default formula Y = 0.299 R + 0.587 G + 0.114 B
    """
    # return cv2.imread(impath, cv2.CV_LOAD_IMAGE_GRAYSCALE)
    return cv2.imread(impath, cv2.IMREAD_GRAYSCALE)


def derive_m(img, rimg):
    """Derive m (intensity) based on paper formula"""

    (rw, cl, ch) = img.shape
    for r in range(rw):
        for c in range(cl):
            rimg[r, c] = int(np.sum(img[r, c]) / 3.0)

    return rimg


def derive_saturation(img, rimg):
    """Derive saturation value for a pixel based on paper formula"""

    s_img = np.array(rimg)
    (r, c) = s_img.shape
    for ri in range(r):
        for ci in range(c):
            # opencv ==> b,g,r order
            s1 = np.clip(img[ri, ci][0] + img[ri, ci][2], 0, 255)
            s2 = 2 * img[ri, ci][1]
            if s1 >= s2:
                s_img[ri, ci] = 1.5 * np.clip((img[ri, ci][2] - rimg[ri, ci]), 0, 255)
            else:
                s_img[ri, ci] = 1.5 * np.clip((rimg[ri, ci] - img[ri, ci][0]), 0, 255)

    return s_img


def check_pixel_specularity(mimg, simg, m_max=0.6, s_max=0.4):
    """Check whether a pixel is part of specular region or not"""

    m_max = np.max(mimg) * m_max
    s_max = np.max(simg) * s_max

    (rw, cl) = simg.shape

    spec_mask = np.zeros((rw, cl), dtype=np.uint8)
    for r in range(rw):
        for c in range(cl):
            if mimg[r, c] >= m_max and simg[r, c] <= s_max:
                spec_mask[r, c] = 255

    return spec_mask


def enlarge_specularity(spec_mask):
    """Use sliding window technique to enlarge specularity
    simply move window over the image if specular pixel detected
    mark center pixel is specular
    win_size = 5x5, step_size = 1
    """

    win_size, step_size = (5, 5), 1
    enlarged_spec = np.array(spec_mask)
    for r in range(0, spec_mask.shape[0], step_size):
        for c in range(0, spec_mask.shape[1], step_size):
            # yield the current window
            win = spec_mask[r : r + win_size[1], c : c + win_size[0]]

            if win.shape[0] == win_size[0] and win.shape[1] == win_size[1]:
                if win[1, 1] != 0:
                    enlarged_spec[r : r + win_size[1], c : c + win_size[0]] = 255 * np.ones((5, 5), dtype=np.uint8)

    return enlarged_spec



# sdcat, Apache-2.0 license
# Filename: cluster/commands.py
# Description:  Clustering commands to cluster both detections and ROIs
import json
import re
import shutil
import sys
import tempfile
import uuid
from datetime import datetime
from pathlib import Path

import click
import ephem
import os
import modin.pandas as pd
import pytz
from PIL import Image
from tqdm import tqdm

from sdcat import __version__ as sdcat_version
from sdcat.cluster.utils import filter_images, combine_csv
from sdcat import common_args
from sdcat.config import config as cfg
from sdcat.logger import info, err, warn
from sdcat.cluster.cluster import cluster_vits


@click.command("detections", help="Cluster detections. See cluster --config-ini to override cluster defaults.")
@common_args.config_ini
@common_args.start_image
@common_args.end_image
@common_args.use_pca
@common_args.use_cuhdbscan
@common_args.skip_visualization
@common_args.alpha
@common_args.cluster_selection_epsilon
@common_args.cluster_selection_method
@common_args.algorithm
@common_args.min_cluster_size
@common_args.min_sample_size
@common_args.vits_batch_size
@common_args.hdbscan_batch_size
@click.option("--det-dir", help="Input folder(s) with raw detection results", multiple=True, required=True)
@click.option("--save-dir", help="Output directory to save clustered detection results", required=True)
@click.option("--device", help="Device to use, e.g. cpu or cuda:0 or cuda to use all cuda devices", type=str, default="cpu")
@click.option("--use-vits", help="Set to using the predictions from the vits cluster model", is_flag=True)
def run_cluster_det(
    det_dir,
    save_dir,
    device,
    use_vits,
    config_ini,
    alpha,
    cluster_selection_epsilon,
    cluster_selection_method,
    algorithm,
    min_cluster_size,
    min_sample_size,
    vits_batch_size,
    hdbscan_batch_size,
    start_image,
    end_image,
    use_pca,
    use_cuhdbscan,
    skip_visualization,
):
    config = cfg.Config(config_ini)
    max_area = int(config("cluster", "max_area"))
    min_area = int(config("cluster", "min_area"))
    min_saliency = int(config("cluster", "min_saliency"))
    alpha = alpha if alpha else float(config("cluster", "alpha"))
    min_cluster_size = min_cluster_size if min_cluster_size else int(config("cluster", "min_cluster_size"))
    min_samples = min_sample_size if min_sample_size else int(config("cluster", "min_samples"))
    cluster_selection_epsilon = (
        cluster_selection_epsilon if cluster_selection_epsilon else float(config("cluster", "cluster_selection_epsilon"))
    )
    cluster_selection_method = (
        cluster_selection_method if cluster_selection_method else config("cluster", "cluster_selection_method")
    )
    algorithm = algorithm if algorithm else config("cluster", "algorithm")
    extract_metadata = config("cluster", "extract_metadata") == "True"
    remove_corners = config("cluster", "remove_corners") == "True"
    remove_bad_images = config("cluster", "remove_bad_images") == "True"
    latitude = float(config("cluster", "latitude"))
    longitude = float(config("cluster", "longitude"))
    min_score = float(config("cluster", "min_score"))
    min_similarity = float(config("cluster", "min_similarity"))
    model = config("cluster", "model")
    allowable_classes = config("cluster", "allowable_classes")
    if len(allowable_classes) > 0:
        allowable_classes = allowable_classes.split(",")

    save_dir = Path(save_dir)
    save_dir.mkdir(parents=True, exist_ok=True)

    csv_files = []
    for d in det_dir:
        info(f"Searching in {d} for detection csv files")
        d_path = Path(d)
        if not d_path.exists():
            err(f"Input path {d} does not exist.")
            return

        csv_files.extend(d_path.rglob("*.csv"))

    info(f"Found {len(csv_files)} detection files in {det_dir}")

    crop_path = save_dir / "crops"
    crop_path.mkdir(parents=True, exist_ok=True)

    tmp_dir = os.getenv("TMPDIR", tempfile.gettempdir())
    with tempfile.TemporaryDirectory(dir=tmp_dir) as temp_dir:
        csv_file = combine_csv(csv_files, Path(temp_dir), crop_path)

        info("Loading detections")
        df = pd.read_csv(csv_file, sep=",")

        if df.empty:
            info(f"No detections found in {det_dir}")
            return

        info(f"Found {len(df)} detections in {det_dir}")

        # Check if the image_path column is empty
        if df["image_path"].isnull().values.any():
            err(f'Found {df["image_path"].isnull().sum()} detections with no image_path')
            return

        # Sort the dataframe by image_path to make sure the images are in order for start_image and end_image filtering
        df = df.sort_values(by="image_path")

        # Remove any rows where the image_path does not exist
        df = df[df["image_path"].apply(lambda x: Path(x).exists())]

        # If start_image is set, find the index of the start_image in the list of images
        if start_image:
            start_image = Path(start_image)
            start_image = start_image.resolve()
            start_image = start_image.stem
            start_image_index = df[df["image_path"].str.contains(start_image)].index[0]
            df = df.iloc[start_image_index:]

        # If end_image is set, find the index of the end_image in the list of images
        if end_image:
            end_image = Path(end_image)
            end_image = end_image.resolve()
            end_image = end_image.stem
            end_image_index = df[df["image_path"].str.contains(end_image)].index[-1]
            df = df.iloc[:end_image_index]

        # Filter the dataframe to only include images in the start_image and end_image range
        df = filter_images(min_area, max_area, min_saliency, min_score, df)

        if df.empty:
            info(
                f"No detections found after filtering min_area={min_area}, max_area={max_area}, min_saliency={min_saliency}, min_score={min_score}"
            )
            return

        # Add in a column for the unique crop name for each detection with a unique id
        # create a unique uuid based on the md5 hash of the box in the row
        df["crop_path"] = df.apply(
            lambda row: f"{row['crop_root']}/{uuid.uuid5(uuid.NAMESPACE_DNS, str(row['x']) + str(row['y']) + str(row['xx']) + str(row['xy']))}.png",
            axis=1,
        )

        # Add in a column for the unique crop name for each detection with a unique id
        df["cluster"] = -1  # -1 is the default value and means that the image is not in a cluster
        df["class"] = "Unknown"
        df["class_s"] = "Unknown"
        df["score"] = 0.0
        df["score_s"] = 0.0

        def within_1_percent_of_corners(row):
            threshold = 0.01  # 1% threshold

            x, y, xx, yy = row["x"], row["y"], row["xx"], row["xy"]

            # Check if any of the coordinates are within 1% of the image corners
            return (
                (0 <= x <= threshold or 1 - threshold <= x <= 1)
                or (0 <= y <= threshold or 1 - threshold <= y <= 1)
                or (0 <= xx <= threshold or 1 - threshold <= xx <= 1)
                or (0 <= yy <= threshold or 1 - threshold <= yy <= 1)
            )

        if remove_corners:
            # Remove any detections that are in any corner of the image
            size_before = len(df)
            df = df[~df.apply(within_1_percent_of_corners, axis=1)]
            size_after = len(df)
            info(f"Removed {size_before - size_after} detections that were in the corners of the image")

        # TODO: refactor this block for modin optimization
        if extract_metadata:
            info("Extracting metadata ...")
            pattern_date0 = re.compile(r"(\d{4})(\d{2})(\d{2})T(\d{2})(\d{2})\.(\d{6})Z")  # 20240923T225833.474105Z
            pattern_date1 = re.compile(r"(\d{4})(\d{2})(\d{2})T(\d{2})(\d{2})(\d{2})Z")  # 20161025T184500Z
            pattern_date2 = re.compile(r"(\d{4})(\d{2})(\d{2})T(\d{2})(\d{2})(\d{2})Z\d*mF*")
            pattern_date3 = re.compile(r"(\d{2})(\d{2})(\d{2})T(\d{2})(\d{2})(\d{2})Z")  # 161025T184500Z
            pattern_date4 = re.compile(r"(\d{2})-(\d{2})-(\d{2})T(\d{2})_(\d{2})_(\d{2})-")  # 16-06-06T16_04_54
            pattern_depth = re.compile(r"_(\d+(?:\.\d+)?)m_")  # 50m_100m_200m_ or _100.0m

            # Grab any additional metadata from the image name, e.g. depth, yearday, number, day/night
            depth = {}
            yearday = {}
            number = {}
            day_flag = {}
            observer = ephem.Observer()
            observer.lat = latitude
            observer.lon = longitude

            def is_day(utc_dt):
                observer.date = utc_dt
                sun = ephem.Sun(observer)
                if sun.alt > 0:
                    return 1
                return 0

            for index, row in tqdm(df.iterrows(), total=len(df), desc="Extracting metadata", unit="image"):
                image_name = Path(row.image_path).name
                if pattern_date0.search(image_name):
                    match = pattern_date0.search(image_name).groups()
                    year, month, day, hour, minute, second, microsecond = map(int, match)
                    dt = datetime(year, month, day, hour, minute, second, microsecond, tzinfo=pytz.utc)
                    yearday[index] = int(dt.strftime("%j")) - 1
                    day_flag[index] = is_day(dt)
                if pattern_date1.search(image_name):
                    match = pattern_date1.search(image_name).groups()
                    year, month, day, hour, minute, second = map(int, match)
                    dt = datetime(year, month, day, hour, minute, second, tzinfo=pytz.utc)
                    yearday[index] = int(dt.strftime("%j")) - 1
                    day_flag[index] = is_day(dt)
                if pattern_date2.search(image_name):
                    match = pattern_date2.search(image_name).groups()
                    year, month, day, hour, minute, second = map(int, match)
                    dt = datetime(year, month, day, hour, minute, second, tzinfo=pytz.utc)
                    yearday[index] = int(dt.strftime("%j")) - 1
                    day_flag[index] = is_day(dt)
                if pattern_date3.search(image_name):
                    match = pattern_date3.search(image_name).groups()
                    year, month, day, hour, minute, second = map(int, match)
                    dt = datetime(year, month, day, hour, minute, second, tzinfo=pytz.utc)
                    yearday[index] = int(dt.strftime("%j")) - 1
                    day_flag[index] = is_day(dt)
                if pattern_date4.search(image_name):
                    match = pattern_date4.search(image_name).groups()
                    year, month, day, hour, minute, second = map(int, match)
                    dt = datetime(year, month, day, hour, minute, second, tzinfo=pytz.utc)
                    yearday[index] = int(dt.strftime("%j")) - 1
                    day_flag[index] = is_day(dt)
                if pattern_depth.search(image_name):
                    match = pattern_depth.match(image_name)
                    depth[index] = float(match.group(1))

            # Add the depth, yearday, day, and night columns to the dataframe if they exist
            if len(depth) > 0:
                df["depth"] = depth
                df["depth"] = df["depth"].astype(int)
            if len(yearday) > 0:
                df["yearday"] = yearday
                df["yearday"] = df["yearday"].astype(int)
            if len(number) > 0:
                df["frame"] = number
            if len(day_flag) > 0:
                df["day"] = day_flag
                df["day"] = df["day"].astype(int)

            # Filter by day/night
            # size_before = len(df)
            # df = df[df['day'] == 1]
            # size_after = len(df)
            # info(f'Removed {size_before - size_after} detections that were at night')

        # Replace any NaNs with 0 and reindex
        df.fillna(0)
        df = df.reset_index(drop=True)

        # Print the first 5 rows of the dataframe
        info(df.head(5))

        if len(df) > 0:
            # Replace / with _ in the model name
            model_machine_friendly = model.replace("/", "_")

            # A prefix for the output files to make sure the output is unique for each execution
            prefix = f'{model_machine_friendly}_{datetime.now().strftime("%Y%m%d_%H%M%S")}'

            # Cluster the detections
            summary = cluster_vits(
                prefix,
                model,
                df,
                save_dir,
                alpha,
                cluster_selection_epsilon,
                cluster_selection_method,
                algorithm,
                min_similarity,
                min_cluster_size,
                min_samples,
                device,
                use_pca=use_pca,
                use_cuhdbscan=use_cuhdbscan,
                use_vits=use_vits,
                skip_visualization=skip_visualization,
                roi=False,
                remove_bad_images=remove_bad_images,
                vits_batch_size=vits_batch_size,
                hdbscan_batch_size=hdbscan_batch_size,
                allowable_classes=allowable_classes,
            )

            if summary is None:
                err("No summary returned from clustering")
                return

            # Add more detail to the summary specific to detections
            summary["sdcat_version"] = sdcat_version
            summary["command"] = " ".join(sys.argv)
            summary["dataset"]["input"] = det_dir
            summary["dataset"]["image_resolution"] = f"{df['image_width'].iloc[0]}x{df['image_height'].iloc[0]} pixels"
            summary["dataset"]["detection_count"] = len(df)

            with open(save_dir / f"{prefix}_summary.json", "w") as f:
                json.dump(summary, f, indent=4)
            info(f"Summary saved to {save_dir / f'{prefix}_summary.json'}")
            info(f"Summary: {json.dumps(summary, indent=4)}")

            # Save a copy of the config.ini file
            save_ini = save_dir / f"{prefix}_config.ini"
            shutil.copy(Path(config_ini), save_ini)
            info(f"Config saved to {save_ini}")
        else:
            warn(f"No detections found to cluster in {det_dir}'")


@click.command("roi", help="Cluster roi. See cluster --config-ini to override cluster defaults.")
@common_args.config_ini
@common_args.use_pca
@common_args.use_cuhdbscan
@common_args.skip_visualization
@common_args.alpha
@common_args.cluster_selection_epsilon
@common_args.cluster_selection_method
@common_args.algorithm
@common_args.min_cluster_size
@common_args.min_sample_size
@common_args.vits_batch_size
@common_args.hdbscan_batch_size
@click.option("--roi-dir", help="Input folder(s) with raw ROI images", multiple=True, required=True)
@click.option("--save-dir", help="Output directory to save clustered detection results", required=True)
@click.option("--device", help="Device to use, e.g. cpu or cuda:0 or cuda to use all cuda devices", type=str)
@click.option("--use-vits", help="Set to using the predictions from the vits cluster model", is_flag=True)
def run_cluster_roi(
    roi_dir,
    save_dir,
    device,
    use_vits,
    config_ini,
    alpha,
    cluster_selection_epsilon,
    cluster_selection_method,
    algorithm,
    min_cluster_size,
    min_sample_size,
    vits_batch_size,
    hdbscan_batch_size,
    use_pca,
    use_cuhdbscan,
    skip_visualization,
):
    config = cfg.Config(config_ini)
    max_area = int(config("cluster", "max_area"))
    min_area = int(config("cluster", "min_area"))
    min_saliency = int(config("cluster", "min_saliency"))
    alpha = alpha if alpha else float(config("cluster", "alpha"))
    min_cluster_size = min_cluster_size if min_cluster_size else int(config("cluster", "min_cluster_size"))
    min_samples = min_sample_size if min_sample_size else int(config("cluster", "min_samples"))
    cluster_selection_epsilon = (
        cluster_selection_epsilon if cluster_selection_epsilon else float(config("cluster", "cluster_selection_epsilon"))
    )
    cluster_selection_method = (
        cluster_selection_method if cluster_selection_method else config("cluster", "cluster_selection_method")
    )
    algorithm = algorithm if algorithm else config("cluster", "algorithm")
    remove_bad_images = config("cluster", "remove_bad_images") == "True"
    min_score = float(config("cluster", "min_score"))
    min_similarity = float(config("cluster", "min_similarity"))
    model = config("cluster", "model")
    allowable_classes = config("cluster", "allowable_classes")
    if len(allowable_classes) > 0:
        allowable_classes = allowable_classes.split(",")

    save_dir = Path(save_dir)
    save_dir.mkdir(parents=True, exist_ok=True)

    # Grab all images from the input directories
    supported_extensions = [".png", ".jpg", ".jpeg", ".JPG", ".JPEG", ".PNG"]
    images = []

    for r in roi_dir:
        roi_path = Path(r)
        for ext in supported_extensions:
            images.extend(list(roi_path.rglob(f"*{ext}")))

    # Create a dataframe to store the combined data in an image_path column in sorted order
    df = pd.DataFrame({"image_path": [str(p) for p in images]})

    info(f"Found {len(df)} detections in {roi_dir}")

    if df.empty:
        info(f"No detections found in {roi_dir}")
        return

    # Sort the dataframe by image_path to make sure the images are in order for start_image and end_image filtering
    df = df.sort_values(by="image_path")

    # Add the image_width and image_height columns to the dataframe
    # Assuming all images are the same size, we can just get the size of the first image and use that for all images
    im_size = Image.open(df["image_path"].iloc[0]).size
    df["image_height"] = im_size[1]
    df["image_width"] = im_size[0]

    # Create a unique crop name for each detection with a unique id
    crop_path = save_dir / "crops"
    crop_path.mkdir(parents=True, exist_ok=True)
    df["crop_path"] = df.apply(lambda row: f'{crop_path}/{Path(row["image_path"]).stem}.png', axis=1)

    # Copy the images to the crop path directory. Images may be cleaned so we want to duplicate them here.
    # Only copy images that do not already exist in the crop path to save time
    to_copy = df[~df["crop_path"].apply(os.path.exists)]
    if to_copy.empty:
        info(f"No images to copy to {crop_path}")
    else:
        for src, dst in tqdm(zip(to_copy["image_path"], to_copy["crop_path"]), total=len(to_copy), desc="Copying images"):
            shutil.copy(src, dst)

    df = filter_images(min_area, max_area, min_saliency, min_score, df)
    if df.empty:
        info(
            f"No detections found after filtering for min_area={min_area}, max_area={max_area}, min_saliency={min_saliency}, min_score={min_score}"
        )
        return

    df["cluster"] = -1  # -1 is the default value and means that the image is not in a cluster
    df["class"] = "Unknown"
    df["class_s"] = "Unknown"
    df["score"] = 0.0
    df["score_s"] = 0.0
    df["x"] = 0.0
    df["y"] = 0.0
    df["xx"] = 1.0
    df["xy"] = 1.0

    # Replace any NaNs with 0 and reindex
    df.fillna(0)
    df = df.reset_index(drop=True)

    # Print the first 5 rows of the dataframe
    info(df.head(5))

    if len(df) > 0:
        model_machine_friendly = model.replace("/", "_")

        # A prefix for the output files to make sure the output is unique for each execution
        prefix = f'{model_machine_friendly}_{datetime.now().strftime("%Y%m%d_%H%M%S")}'

        # Cluster the ROIs
        summary = cluster_vits(
            prefix,
            model,
            df,
            save_dir,
            alpha,
            cluster_selection_epsilon,
            cluster_selection_method,
            algorithm,
            min_similarity,
            min_cluster_size,
            min_samples,
            device,
            use_pca=use_pca,
            use_cuhdbscan=use_cuhdbscan,
            use_vits=use_vits,
            skip_visualization=skip_visualization,
            roi=True,
            remove_bad_images=remove_bad_images,
            vits_batch_size=vits_batch_size,
            hdbscan_batch_size=hdbscan_batch_size,
            allowable_classes=allowable_classes,
        )

        # Add more detail to the summary specific to ROIs
        summary["sdcat_version"] = sdcat_version
        summary["command"] = " ".join(sys.argv)
        summary["dataset"]["roi"] = True
        summary["dataset"]["input"] = roi_dir
        summary["dataset"]["image_resolution"] = "224x224 pixels"
        summary["dataset"]["detection_count"] = len(df)

        with open(save_dir / f"{prefix}_summary.json", "w") as f:
            json.dump(summary, f, indent=4)
        info(f'Summary saved to {save_dir / f"{prefix}_summary.json"}')
        info(f"Summary: {json.dumps(summary, indent=4)}")

        # Save a copy of the config.ini file
        shutil.copy(Path(config_ini), save_dir / f"{prefix}_config.ini")
        info(f'Config saved to {save_dir / f"{prefix}_config.ini"}')
    else:
        warn(f"No images found to cluster in {roi_dir}'")



# sdcat, Apache-2.0 license
# Filename: sdcat/detect/commands.py
# Description:  Command line interface for running detection on images using SAHI and saliency detection algorithms.
from concurrent.futures import ProcessPoolExecutor
import os
import shutil
from pathlib import Path

import click
from tqdm import tqdm
import pandas as pd
import torch

from sdcat import common_args
from sdcat.config import config as cfg
from sdcat.config.config import default_config_ini
from sdcat.detect.filter_util import process_image
from sdcat.detect.model_util import create_model
from sdcat.detect.sahi_detector import run_sahi_detect_bulk, run_sahi_detect
from sdcat.detect.saliency_detector import run_saliency_detect, run_saliency_detect_bulk
from sdcat.logger import info, warn, create_logger_file

default_model = "MBARI-org/megamidwater"


@click.command(
    "detect",
    help=f"Detect objects in images.  By default, runs both SAHI and saliency blob detection and combines using NMS. "
    f"Use --skip-sahi or --skip-saliency to exclude. "
    f"See --config-ini to override detection defaults in {default_config_ini}.",
)
@common_args.config_ini
@common_args.start_image
@common_args.end_image
@click.option("--show", is_flag=True, help="Show algorithm steps.")
@click.option("--image-dir", required=True, help="Directory with images to run sliced detection.")
@click.option("--save-dir", required=True, help="Save detections to this directory.")
@click.option("--save-roi", is_flag=True, help="Save each region of interest/detection.")
@click.option("--roi-size", type=int, default=224, help="Rescale the region of interest.")
@click.option("--device", default="cpu", help="Device to use, e.g. cpu or cuda:0")
@click.option(
    "--spec-remove",
    is_flag=True,
    help="Run specularity removal algorithm on the images before processing. "
    "**CAUTION**this is slow. Set --scale-percent to < 100 to speed-up",
)
@click.option("--skip-sahi", is_flag=True, help="Skip sahi detection.")
@click.option("--skip-saliency", is_flag=True, help="Skip saliency detection.")
@click.option("--conf", default=0.1, type=float, help="Confidence threshold.")
@click.option("--scale-percent", default=80, type=int, help="Scaling factor to rescale the images before processing.")
@click.option("--model", default=default_model, help=f"Model to use. Defaults to {default_model}")
@click.option("--model-type", help="Type of model, e.g. yolov5, yolov8. Defaults to auto-detect.")
@click.option("--slice-size-width", type=int, help="Slice width size, leave blank for auto slicing")
@click.option("--slice-size-height", type=int, help="Slice height size, leave blank for auto slicing")
@click.option(
    "--postprocess-match-metric",
    default="IOS",
    help="Postprocess match metric for NMS. postprocess_match_metric IOU for intersection over union, IOS for intersection over smaller area.",
)
@click.option("--overlap-width-ratio", type=float, default=0.4, help="Overlap width ratio for NMS")
@click.option("--overlap-height-ratio", type=float, default=0.4, help="Overlap height ratio for NMS")
@click.option(
    "--clahe",
    is_flag=True,
    help="Run the CLAHE algorithm to contrast enhance before detection useful images with non-uniform lighting",
)
def run_detect(
    show: bool,
    image_dir: str,
    save_dir: str,
    save_roi: bool,
    roi_size: int,
    model: str,
    model_type: str,
    slice_size_width: int,
    slice_size_height: int,
    scale_percent: int,
    postprocess_match_metric: str,
    overlap_width_ratio: float,
    overlap_height_ratio: float,
    device: str,
    conf: float,
    skip_sahi: bool,
    skip_saliency: bool,
    spec_remove: bool,
    config_ini: str,
    clahe: bool,
    start_image: str,
    end_image: str,
):
    config = cfg.Config(config_ini)
    clahe = clahe if clahe else config("detect", "clahe") == "True"
    block_size = int(config("detect", "block_size"))
    min_std = float(config("detect", "min_std"))
    max_area = int(config("detect", "max_area"))
    min_area = int(config("detect", "min_area"))
    min_saliency = int(config("detect", "min_saliency"))
    class_agnostic = config("detect", "class_agnostic") == "True"
    allowable_classes = []
    if not class_agnostic:
        allowable_classes = config("detect", "allowable_classes")
        if len(allowable_classes) > 0:
            allowable_classes = allowable_classes.split(",")

    create_logger_file("detect")

    if not skip_sahi:
        if "cuda" in device:
            torch.cuda.empty_cache()
            num_devices = torch.cuda.device_count()
            info(f"{num_devices} cuda devices available")
            device_ = torch.device(device)
            torch.cuda.set_device(device_)
            device = "cuda"
        detection_model = create_model(model, conf, device, model_type)

    if Path(model).is_dir():
        model = Path(model).name

    images_path = Path(image_dir)
    base_path = Path(save_dir) / model

    # clean-up any previous results
    if base_path.exists():
        info(f"Removing {base_path}")
        shutil.rmtree(base_path)

    if not skip_sahi:
        save_path_base = Path(save_dir) / model
    else:
        save_path_base = Path(save_dir)

    save_path_det_raw = save_path_base / "det_raw" / "csv"
    save_path_det_filtered = save_path_base / "det_filtered" / "csv"
    save_path_det_roi = save_path_base / "det_filtered" / "crops"
    save_path_viz = save_path_base / "vizresults"

    if save_roi:
        save_path_det_roi.mkdir(parents=True, exist_ok=True)
        for f in save_path_det_roi.rglob("*"):
            os.remove(f)
    save_path_det_raw.mkdir(parents=True, exist_ok=True)
    save_path_det_filtered.mkdir(parents=True, exist_ok=True)
    save_path_viz.mkdir(parents=True, exist_ok=True)

    # Run on all images recursively
    images = [
        file
        for file in sorted(images_path.rglob("*"))
        if file.as_posix().endswith(("jpeg", "png", "jpg", "JPEG", "PNG", "JPG", "tif", "tiff"))
    ]

    if len(images) == 0:
        warn(f"No images found in {images_path}")
        return

    start_index = 0
    end_index = len(images)
    # If start_image is set, find the index of the start_image in the list of images
    if start_image:
        start_image = Path(start_image)
        start_index = next((i for i, image in enumerate(images) if start_image.name in image.name), None)
        if start_index is None:
            warn(f"Start image {start_image} not found in images")
            return

    # If end_image is set, find the index of the end_image in the list of images
    if end_image:
        end_image = Path(end_image)
        end_index = next((i for i, image in enumerate(images) if end_image.name in image.name), None)
        if end_index is None:
            warn(f"End image {end_image} not found in images")
            return

    images = images[start_index : end_index + 1]
    num_images = len(images)
    info(f"Found {num_images} images in {images_path}")

    if num_images == 0:
        return

    num_processes = min(os.cpu_count(), num_images)
    if not skip_saliency:
        images_per_process = min(num_images, 2)  # Process 2 images at a time for saliency detection
        info(f"Using {num_processes} processes to compute {num_images} images {images_per_process} at a time ...")
        args = [
            (
                spec_remove,
                scale_percent,
                images[i : i + images_per_process],
                save_path_det_raw,
                min_std,
                block_size,
                clahe,
                show,
            )
            for i in range(0, num_images, images_per_process)
        ]
        with ProcessPoolExecutor(max_workers=num_processes) as executor:
            futures = [executor.submit(run_saliency_detect_bulk, *arg) for arg in args]
            for future in futures:
                future.result()  # Ensure any exceptions are raised

    if device == "cpu":
        if not skip_sahi:
            args = [
                (
                    scale_percent,
                    slice_size_width,
                    slice_size_height,
                    [image],
                    save_path_det_raw,
                    detection_model,
                    postprocess_match_metric,
                    overlap_width_ratio,
                    overlap_height_ratio,
                    allowable_classes,
                    class_agnostic,
                )
                for image in images
            ]
            with ProcessPoolExecutor(max_workers=num_processes) as executor:
                futures = [executor.submit(run_sahi_detect_bulk, *arg) for arg in args]
                for future in futures:
                    future.result()
    else:
        for f in tqdm(images):
            if not skip_saliency:
                run_saliency_detect(
                    spec_remove,
                    scale_percent,
                    f.as_posix(),
                    (save_path_det_raw / f"{f.stem}.csv").as_posix(),
                    None,
                    min_std,
                    block_size,
                    clahe,
                    show,
                )
            if not skip_sahi:
                run_sahi_detect(
                    scale_percent,
                    slice_size_width,
                    slice_size_height,
                    f,
                    (save_path_det_raw / f"{f.stem}.sahi.csv"),
                    detection_model,
                    postprocess_match_metric,
                    overlap_width_ratio,
                    overlap_height_ratio,
                    allowable_classes,
                    class_agnostic,
                )

    # Count the number of detections in all the images
    total_detections = sum(pd.read_csv(f).shape[0] for f in save_path_det_raw.rglob("*.csv"))
    total_images = len(images)

    args = [
        (
            image,
            save_path_base,
            save_path_det_raw,
            save_path_det_filtered,
            save_path_det_roi,
            save_path_viz,
            min_area,
            max_area,
            min_saliency,
            class_agnostic,
            save_roi,
            roi_size,
        )
        for image in images
    ]

    # Process images in parallel
    num_processes = min(os.cpu_count(), total_images)
    with ProcessPoolExecutor(max_workers=num_processes) as executor:
        futures = [executor.submit(process_image, *arg) for arg in args]
        results = [future.result() for future in futures]  # raises exceptions if any

    total_filtered = sum(results)
    info(f"Found {total_detections} total localizations in {total_images} images with {total_filtered} after NMS")
    info("Done")



# sdcat, Apache-2.0 license
# Filename: common_args.py
# Description: Common arguments for processing commands

import click

from sdcat.config.config import default_config_ini

# Common arguments for processing commands
config_ini = click.option(
    "--config-ini",
    type=str,
    default=default_config_ini,
    help=f"Path to config file to override. Defaults are in {default_config_ini}. Copy to your own custom.ini file to override",
)

start_image = click.option("--start-image", type=str, help="Start image name")

end_image = click.option("--end-image", type=str, help="End image name")

alpha = click.option(
    "--alpha",
    type=float,
    help="Alpha is a parameter that controls the linkage. See https://hdbscan.readthedocs.io/en/latest/parameter_selection.html. "
    "Default is 0.92. Increase for less conservative clustering, e.g. 1.0",
)

cluster_selection_epsilon = click.option(
    "--cluster-selection-epsilon",
    type=float,
    help="Epsilon is a parameter that controls the linkage. " "Default is 0. Increase for less conservative clustering",
)

cluster_selection_method = click.option(
    "--cluster-selection-method",
    type=str,
    help="Method for selecting the optimal number of clusters. " "Default is leaf. Options are leaf, eom, and dill",
)

algorithm = click.option(
    "--algorithm", type=str, help="Algorithm for clustering. Default is best.  best, generic, prims_kdtree, boruvka_kdtree"
)

min_cluster_size = click.option(
    "--min-cluster-size",
    type=int,
    help="The minimum number of samples in a group for that group to be considered a cluster. "
    "Default is 2. Increase for less conservative clustering, e.g. 5, 15",
)

min_sample_size = click.option(
    "--min-sample-size",
    type=int,
    help="The number of samples in a neighborhood for a point to be considered as a core point. "
    "This includes the point itself. Default is 1. Increase for more conservative clustering",
)

vits_batch_size = click.option(
    "--vits-batch-size", type=int, default=32, help="Batch size for processing images. Default is 32"
)

hdbscan_batch_size = click.option(
    "--hdbscan-batch-size",
    type=int,
    default=50000,
    help="Batch size for HDBSCAN. Default is 50000. Increase for your available CPU/GPU memory. ",
)

use_pca = click.option(
    "--use-pca", is_flag=True, help="Use PCA for dimensionality reduction before clustering. Default is False"
)

use_cuhdbscan = click.option(
    "--use-cuhdbscan",
    is_flag=True,
    help="Use cuHDBSCAN for clustering. Default is False. Requires a GPU with CUDA support. Faster than hdbscan but not as accurate as it uses euclidean distance instead of cosine distance",
)

skip_visualization = click.option("--skip-visualization", is_flag=True, help="Skip visualization. Default is False")



# sdcat, Apache-2.0 license
# Filename: sdcat/ml/saliency.py
# Description:  Miscellaneous saliency functions for detecting targets in images.
import os
import tempfile
from concurrent.futures import ProcessPoolExecutor

import cv2
import numpy as np
import torch
import pandas as pd
from sahi.postprocess.combine import nms

from sdcat.cluster.utils import combine_csv
from sdcat.logger import debug, info
from pathlib import Path

save = True


def process_contour(
    out_csv_file: str, contours: np.ndarray, gray: np.ndarray, min_blob_area: int, max_blob_area: int, scale: float = 1.0
):
    """
    Process a single contour. Save the results to a csv file.
    This is a separate function, so it can be run in parallel.
    :param out_csv_file: The csv path to save the results
    :param contours:  List of contours
    :param gray: Grayscale image to use for extracting basic statistics
    :param min_blob_area: The minimum blob area to consider
    :param max_blob_area: The maximum blob area to consider
    :param scale: The scale factor used to rescale the image
    :return:
    """
    df = pd.DataFrame()
    for c in contours:
        area = cv2.contourArea(c)
        x, y, w, h = cv2.boundingRect(c)
        mask = np.zeros(gray.shape, np.uint8)

        if min_blob_area < area < max_blob_area:
            cv2.drawContours(mask, [c], 0, 255, -1)
            masked_image = cv2.bitwise_and(gray, gray, mask=mask)
            mean_intensity = cv2.mean(masked_image, mask=mask)[0]
            max_intensity = np.max(masked_image[mask != 0])
            std_intensity = np.std(masked_image[mask != 0])

            # Small blobs that are very bright are likely reflections, so ignore them
            if max_intensity > 200.0 and area < 10000 and std_intensity < 30.0:
                debug(
                    f"Skipping possible reflection blob: {area}, mean: {mean_intensity}, max: {max_intensity}, std: {std_intensity}, area: {area}"
                )
                continue
            if std_intensity > 8.0:
                debug(
                    f"Found blob area: {area}, mean: {mean_intensity}, max: {max_intensity}, std: {std_intensity}, area: {area}"
                )
                df = pd.concat(
                    [
                        df,
                        pd.DataFrame(
                            {
                                "image_path": "",
                                "class": "Unknown",
                                "score": 0.1,
                                "area": area,
                                "mean_intensity": mean_intensity,
                                "std_intensity": std_intensity,
                                "x": int(x / scale),
                                "y": int(y / scale),
                                "xx": int((x + w) / scale),
                                "xy": int((y + h) / scale),
                                "w": int(w / scale),
                                "h": int(h / scale),
                            },
                            index=[0],
                        ),
                    ]
                )

    if len(df) > 0:
        # Save the df to a csv file for later processing
        df.to_csv(out_csv_file, mode="w", header=True, index=False)


def homomorphic_filtering(image_in: np.ndarray, alpha: float = 0.0003125, beta: float = 1.00):
    """
    Apply homomorphic filtering to an image to remove illumination artifacts
    :param image_in:
    :param alpha:
    :param beta:
    :return:
    """
    img_float32 = np.float32(image_in)
    log_image = np.log1p(img_float32)
    fft = np.fft.fft2(log_image)
    rows, cols = img_float32.shape
    crow, ccol = rows // 2, cols // 2
    mask = np.ones((rows, cols), np.uint8)
    r = 30
    center = (crow, ccol)
    x, y = np.ogrid[:rows, :cols]
    mask_area = (x - center[0]) ** 2 + (y - center[1]) ** 2 <= r * r
    mask[mask_area] = 0
    fft_shift = fft * mask
    img_back = np.fft.ifft2(fft_shift)
    img_back = np.exp(np.real(img_back))
    img_filtered = (img_back - np.min(img_back)) / (np.max(img_back) - np.min(img_back))
    img_filtered = np.power(img_filtered, alpha) * beta
    img_filtered = np.uint8(img_filtered * 255)
    return img_filtered


def extract_blobs(saliency_map: np.ndarray, img_gray: np.ndarray, img_color: np.ndarray, show=False) -> pd.DataFrame:
    """
    Extract blobs from a saliency map
    :param saliency_map: normalized saliency map
    :param img_gray: grayscale image used to extract basic statistics
    :param img_color: color image used when showing the results
    :param show: True to show the results
    :return: pandas dataframe of blobs
    """

    # Find contours in the salient object mask
    contours, _ = cv2.findContours(saliency_map, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_TC89_L1)

    # Show the contours on the image
    if show:
        result_image = img_color.copy()
        cv2.drawContours(result_image, contours, -1, (0, 255, 0), 2)
        cv2.imshow(f"{len(contours)} Contours", result_image)
        cv2.waitKey(0)

    # Iterate through the contours and filter out by size and standard deviation of intensity
    min_blob_area = 10
    max_blob_area = int(img_gray.shape[0] * img_gray.shape[1] / 2)  # 50% of the image area for the largest blob

    # Determine number of processes
    num_processes = min(os.cpu_count(), len(contours) // 100)
    num_processes = max(1, num_processes)
    info(f"Using {num_processes} processes to compute {len(contours)} 100 at a time ...")

    # Work in a temporary directory
    tmp_dir = os.getenv("TMPDIR", tempfile.gettempdir())
    with tempfile.TemporaryDirectory(dir=tmp_dir) as temp_path:
        temp_path = Path(temp_path)
        gray = img_gray

        # Prepare arguments
        args = [
            ((temp_path / f"{i}_det.csv").as_posix(), contours[i : i + 100], gray, min_blob_area, max_blob_area)
            for i in range(0, len(contours), 100)
        ]

        # Run in parallel
        with ProcessPoolExecutor(max_workers=num_processes) as executor:
            futures = [executor.submit(process_contour, *arg) for arg in args]
            for future in futures:
                future.result()  # Re-raises any exception that occurred

        # Combine the results
        info(f"Combining blob detection results in {temp_path}")
        csv_files = list(temp_path.glob("*.csv"))
        csv_file = combine_csv(csv_files, temp_path / "det.csv")
        df = pd.read_csv(csv_file, sep=",")

        # Remove the temporary files
        for f in csv_files:
            f.unlink()

    if show:
        result_image = img_color.copy()
        for i, row in df.iterrows():
            x = row["x"]
            y = row["y"]
            w = row["w"]
            h = row["h"]
            std_intensity = row["std_intensity"]
            cv2.rectangle(result_image, (x, y), (x + w, y + h), (81, 12, 51), 2)
            std_intensity_int = np.round(std_intensity * 100)
            cv2.putText(result_image, f"{std_intensity_int}", (x, y), cv2.FONT_HERSHEY_SIMPLEX, 0.5, (255, 255, 255), 2)

        cv2.imshow(f"{len(df)} Bounding Boxes around Salient Regions", result_image)
        cv2.waitKey(0)
        cv2.destroyAllWindows()

    return df


def fine_grained_saliency_pyramid(image_path: str, show=False) -> np.ndarray:
    """
    Compute a fine-grained saliency map and normalize it
    :param image_path:  path to the image
    :param show:  True to show the results
    :return: normalized saliency map
    """
    # Load the original image
    image = cv2.imread(image_path)

    img_hsv = cv2.cvtColor(image, cv2.COLOR_RGB2HSV)

    # Blur the saturation channel to remove noise
    img_hsv[:, :, 1] = cv2.GaussianBlur(img_hsv[:, :, 1], (15, 15), 0)

    # Build an image pyramid, starting with the original image, then downsampling it 2 times
    pyramid = [img_hsv]
    for i in range(1, 3):
        pyramid.append(cv2.pyrDown(pyramid[-1]))

    # Initialize an empty saliency map
    saliency_map = np.zeros_like(image[:, :, 0], dtype=np.float32)

    # Process each level of the pyramid
    for level, img in enumerate(pyramid):
        # Extract the hue, saturation, and value channels
        hue = img[:, :, 0]
        saturation = img[:, :, 1]
        value = img[:, :, 2]

        # Run homomorphic filtering on each channel
        filtered_hue = homomorphic_filtering(hue)
        filtered_saturation = saturation
        filtered_value = homomorphic_filtering(value)

        blurred_hue = cv2.GaussianBlur(filtered_hue, (3, 3), 0)
        blurred_saturation = cv2.GaussianBlur(filtered_saturation, (5, 5), 0)
        blurred_value = cv2.GaussianBlur(filtered_value, (3, 3), 0)

        # Calculate the center and surround regions for each channel
        center_hue = cv2.GaussianBlur(blurred_hue, (0, 0), 2)
        surround_hue = blurred_hue - center_hue

        center_saturation = cv2.GaussianBlur(blurred_saturation, (3, 3), 2)
        surround_saturation = blurred_saturation - center_saturation

        center_value = cv2.GaussianBlur(blurred_value, (0, 0), 2)
        surround_value = blurred_value - center_value

        # Normalize the values to the range [0, 255]
        center_hue = cv2.normalize(center_hue, None, 0, 255, cv2.NORM_MINMAX)
        surround_hue = cv2.normalize(surround_hue, None, 0, 255, cv2.NORM_MINMAX)

        center_saturation = cv2.normalize(center_saturation, None, 0, 255, cv2.NORM_MINMAX)
        surround_saturation = cv2.normalize(surround_saturation, None, 0, 255, cv2.NORM_MINMAX)

        center_value = cv2.normalize(center_value, None, 0, 255, cv2.NORM_MINMAX)
        surround_value = cv2.normalize(surround_value, None, 0, 255, cv2.NORM_MINMAX)

        # Combine the center and surround regions for each channel
        # Reduce the weight of the saturation channel
        saliency_hue = (center_hue - surround_hue) * (3**level)
        saliency_saturation = (center_saturation - surround_saturation) / (3**level)
        saliency_value = (center_value - surround_value) * (3**level)

        # Combine the saliency maps into the final saliency map
        saliency_level = (saliency_hue + saliency_saturation + saliency_value) / 3

        # Resize the saliency map to the original image size
        saliency_level = cv2.resize(saliency_level, (image.shape[1], image.shape[0]))

        # Accumulate the saliency map at each level
        saliency_map += saliency_level

    # Normalize the final saliency map
    saliency_map = cv2.normalize(saliency_map, None, 0, 255, cv2.NORM_MINMAX)

    # Blur the saliency map to remove noise
    saliency_map = cv2.GaussianBlur(saliency_map, (15, 15), 0)

    if show:
        # Display the original image and the saliency map
        cv2.imshow("Original Image", image)
        cv2.imshow("Fine-Grained Saliency Map", saliency_map.astype(np.uint8))
        if save:
            cv2.imwrite("my_saliency_map.jpg", saliency_map.astype(np.uint8))
        cv2.waitKey(0)

    # Threshold the saliency map
    mean, std = cv2.meanStdDev(saliency_map)
    info(f"Mean: {mean}, Std: {std}")
    _, saliency_map = cv2.threshold(saliency_map.astype(np.uint8), int(mean - 1.5 * std), 255, cv2.ADAPTIVE_THRESH_MEAN_C)

    if show:
        # Display the thresholded saliency map
        cv2.imshow("Thresholded Saliency Map", saliency_map.astype(np.uint8))
        cv2.waitKey(0)

    # Invert the saliency map so that the salient regions are white
    saliency_map = cv2.bitwise_not(saliency_map)

    # Connect the salient regions
    kernel = np.ones((5, 5), np.uint8)
    saliency_map = cv2.dilate(saliency_map, kernel, iterations=5)
    return saliency_map


if __name__ == "__main__":
    # Read all images from the tests/data/kelpflow directory
    # Get the path of the this file
    # test_path = Path(__file__).parent.parent.parent / 'tests' / 'data' / 'bird'
    # test_path = Path(__file__).parent.parent.parent / 'tests' / 'data' / 'whale'
    # test_path = Path(__file__).parent.parent.parent / 'tests' / 'data' / 'kelpflow'
    # test_path = Path( __file__).parent.parent.parent / 'tests' / 'data' / 'glare'
    # test_path = Path(__file__).parent.parent.parent / 'tests' / 'data' / 'dolphin'
    # test_path = Path(__file__).parent.parent.parent / 'tests' / 'data' / 'all'
    test_path = Path(__file__).parent.parent.parent / "tests" / "data" / "jelly_DSC"
    show = True
    for image_path in test_path.glob("*.JPG"):
        saliency_map = fine_grained_saliency_pyramid(image_path.as_posix())
        img_gray = cv2.cvtColor(cv2.imread(image_path.as_posix()), cv2.COLOR_RGB2GRAY)
        img_color = cv2.imread(image_path.as_posix())
        df = extract_blobs(saliency_map, img_gray, img_color, show=show)

        print(f"Found {len(df)} blobs in {image_path}")

        # Convert the x, y, xx, xy, score columns to a list of Tensors that is Shape: [num_boxes,5].
        pred_list = df[["x", "y", "xx", "xy", "score"]].values.tolist()
        pred_list = torch.tensor(pred_list)  # convert the list of Tensors to a list of Tensors
        print(f"Running NMS on {len(df)} predictions")
        nms_pred_idx = nms(pred_list, "IOU", 0.1)  # run nms with 0.1 IOU
        nms_pred_list = pred_list[nms_pred_idx].tolist()
        df_all = pd.DataFrame(nms_pred_list, columns=["x", "y", "xx", "xy", "score"])
        print(f"{len(df_all)} predictions found after NMS")

        # Plot boxes on the input frame
        pred_list = df_all[["x", "y", "xx", "xy", "score"]].values.tolist()
        for p in pred_list:
            # Draw a rectangle with red line borders of thickness of 3 px
            image_raw = cv2.rectangle(img_color, (int(p[0]), int(p[1])), (int(p[2]), int(p[3])), (0, 0, 255), 3)

        cv2.imshow("final", img_color)
        cv2.waitKey(0)
        cv2.destroyAllWindows()
